/*********************************************************************

	RosettaViewDemo.c

	Alphabet Rosetta

	(c)Copyright 2017-2019  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

********************************************************************

	This is a Rosetta view library demo for any OS
		- It reads Rosetta, SIE and XBRL/GL-files
			* Rosetta specification (www.alphabet.se)
			* SIE4b specification (www.sie.se)
			* Rosetta XBRL/GL Taxonomy specification (www.alphabet.se)
		- It writes the files specified in the prompt command

	Content of RosettaViewDemo.c
	1.) main function
	2.) SaveWorkFileFilePathName
	3.) Locale functions (to support the RosettaView functions)

	In Copyright.h there are the feature settings.

*********************************************************************/
#define FIO_DEMO 1
//$(SolutionDir)\..; $(SolutionDir)\..\RosettaFio;
#include "RosettaFioTypes.h"
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <stddef.h>
#include <locale.h>

#define COUNTRY_ISO3166 "US"
#define LANGUAGE_ISO639 "en"
#define MAX_PATH          260
static BOOL SaveWorkFileFilePathName(InstanceDataType *pInstanceData, char *pFilePathName, BITMASK bmCommand, LocaleDataType *pLocaleData);
static LocaleDataType *GetLocales(LocaleDataType *pLocaleData);
char *GetSystemUser(char **ppTarget);
static char *GetLocaleThousandGrouping(char *grouping);
/********************************************************************

main(int argc, char **argv)

*********************************************************************/
main(int argc, char **argv, char *envp[])
{
	char *pOpenDataFilePathName = 0;
	char *pSaveDataFilePathName = 0;
	BITMASK bmSave = 0;
	AccountingType *pAccounting = 0;
	size_t ztInstanceData = 0;
	LocaleDataType *pLocaleData = GetLocales(0);
	{
		char *pEnv = 0;
		int i = 0;
		while (envp[i++])
			pEnv = envp[i];
	}
	if (argc)
	{
		if (argc == 1) // Nothing
			printf(DEMO_HELP_TEXT);
		else if (argc > 1) {
			if ((argv[1][0] == '?') || (argv[1][1] == '?')) {
				printf(DEMO_HELP_TEXT);
			}
			else {
				pOpenDataFilePathName = argv[1];
				printf("\n\nOpening %s\n", pOpenDataFilePathName);
				char *pBuffer = FileRead(pOpenDataFilePathName);
				if (pBuffer) {
					printf("Reading Success! \n");
					ztInstanceData = InterpretFile(
						pBuffer,
						pOpenDataFilePathName,
						pLocaleData);
					if ((((int)ztInstanceData) > 0) || (((int)ztInstanceData) < (-100))) {
						InstanceDataType *pInstanceData = (InstanceDataType *)ztInstanceData;
						printf("Interpreting success!\n");
						SumUpAccounting(pInstanceData->vpWorkData);
						SumUpBalanceTotals(pInstanceData->vpWorkData);
						Sort(pInstanceData->vpWorkData, VOUCHERS_BY_DATE);
						pInstanceData->pLocaleData = pLocaleData;
						pAccounting = pInstanceData->vpWorkData;
						if (argc > 1) {
							int i = 2;
							for (i = 2; i < argc; i++) {
								if (*argv[i] == '-') {
									switch (argv[i][1]) {
#if(PROGRAM_FEATURES_MASK & PROGRAM_FEATURES_XGL)
									case 'x':
									case 'X': // XBRL/GL
										switch (argv[i][2]) {
										case 'b':
										case 'B':
											bmSave = FILE_TYPE_XGL | CONTENT_BALANCE;
											break;
										case 'l':
										case 'L':
											bmSave = FILE_TYPE_XGL | CONTENT_LEDGER;
											break;
										case 'j':
										case 'J':
											bmSave = FILE_TYPE_XGL | CONTENT_VOUCHERS;
											break;
										default:
											break;
										}
										pSaveDataFilePathName = CopyString(0, &argv[i][3]);
										break;
#endif
#if(PROGRAM_FEATURES_MASK & PROGRAM_FEATURES_SIE)
									case 's':
									case 'S': // SIE
										switch (argv[i][2]) {
										case 'b':
										case 'B':
											bmSave = FILE_TYPE_SIE | CONTENT_BALANCE;
											break;
										case 'l':
										case 'L':
											bmSave = FILE_TYPE_SIE | CONTENT_LEDGER;
											break;
										case 'j':
										case 'J':
											bmSave = FILE_TYPE_SIE | CONTENT_VOUCHERS;
											break;
										default:
											break;
										}
										pSaveDataFilePathName = CopyString(0, &argv[i][3]);
										break;
#endif
#if(PROGRAM_FEATURES_MASK & PROGRAM_FEATURES_ROSETTA)
									case 'r':
									case 'R': // Rosetta
										switch (argv[i][2]) {
										case 'b':
										case 'B':
											bmSave = FILE_TYPE_STD | CONTENT_BALANCE;
											break;
										case 'l':
										case 'L':
											bmSave = FILE_TYPE_STD | CONTENT_LEDGER;
											break;
										case 'j':
										case 'J':
											bmSave = FILE_TYPE_STD | CONTENT_VOUCHERS;
											break;
										default:
											break;
										}
										pSaveDataFilePathName = CopyString(0, &argv[i][3]);
										break;
#endif
									default: // Error
										printf("\nUnrecognised Command\n\n");
										break;
									}
								}
								if (bmSave) {
									if (!(pSaveDataFilePathName && *pSaveDataFilePathName)) {
										// Set file name if missing
										pSaveDataFilePathName = calloc(1024, sizeof(char));
										strcpy(pSaveDataFilePathName, pOpenDataFilePathName);
										CreateGeneralTitle(pInstanceData, pSaveDataFilePathName, bmSave, 0, 0, 0, MAX_PATH);
									}
									printf("Saving %s\n", pSaveDataFilePathName);
									if (SaveWorkFileFilePathName(pInstanceData, pSaveDataFilePathName, bmSave, pLocaleData))
										printf("Saving Failure! \n\n");
									else
										printf("Saving Success! \n\n");

									// Reset
									free(pSaveDataFilePathName);
									pSaveDataFilePathName = 0;
									bmSave = 0;
								}
							}
						}
						DestroyInstanceData(pInstanceData);
					}
					else
						printf("Interpreting Failure!\n");
				}
				else
					printf("Reading Failure! \n\n");
			}
		}
	}
	pLocaleData = DestroyLocales(pLocaleData);
	return 0;
}

/********************************************************************

static size_t SaveWorkFileFilePathName(

*********************************************************************/
static BOOL SaveWorkFileFilePathName(
	InstanceDataType *pInstanceData,
	char *pFilePathName,
	BITMASK bmCommand,
	LocaleDataType *pLocaleData)
{
	if ((bmCommand) && (pFilePathName && *pFilePathName))
	{
		CheckActive(
			pInstanceData->vpWorkData,
			0,
			0,
			0,
			FALSE,
			0);

		switch (bmCommand & FILE_TYPE_MASK) {
		case FILE_TYPE_SIE:
			return FileWriteSie(
				pFilePathName,
				pInstanceData->vpWorkData,
				AppGetContentFromBm(bmCommand),
				pLocaleData->pOperator, //char *pGenUserName,
				TRUE, // BOOL CompactFileOption,
				FALSE, // LeanChartsOption,
				TRUE, // QuantityOption,
				TRUE, // PeriodOption,
				TRUE, // ObjectOption,
				TRUE, // AccountSortOption,
				TRUE, // DimensionSortOption,
				TRUE, // ObjectSortOption,
				TRUE, // VoucherSortOption,
				0, // pInitialDate,
				0, // pFinalDate,
				0, // FilterEnteredBy,
				BRAND_TEXT_LONG,
				PROGRAM_VERSION,
				pLocaleData->pLocaleLanguage,
				FALSE);
		case FILE_TYPE_XGL:
			return FileWriteXbrlGl(
				pFilePathName,
				pInstanceData->vpWorkData,
				AppGetContentFromBm(bmCommand),
				pLocaleData->pOperator,
				0, // Stylesheet 
				TRUE, // ChartsOption,
				TRUE, // ObjectOption,
				TRUE, // BalanceOption,
				TRUE, // PeriodOption,
				TRUE, // VouchersOption,
				TRUE, // QuantityOption,
				TRUE, // LeanChartsOption,
				TRUE, // CompactFileOption,
				TRUE, // ExtendedOption,
				0, // pInitialDate,
				0, // pFinalDate,
				0, // FilterEnteredBy,
				BRAND_TEXT_LONG,
				PROGRAM_VERSION,
				pLocaleData);
		case FILE_TYPE_STD:
			return WriteWorkData(pInstanceData->vpWorkData, AppGetContentFromBm(bmCommand), pFilePathName);
			break;
		default:
			return 0;
		}
	}
	return 0;
}

/********************************************************************

Locale functions

*********************************************************************/
/********************************************************************

static LocaleDataType *GetLocales(

*********************************************************************/
static LocaleDataType *GetLocales(
	LocaleDataType *pLocaleData)
{
	// Note that the ANSI C Locale functions are not implemented in
	// Java NDK and must be checked in other environment of choice. 
	char *pLocale = setlocale(LC_ALL, "");
	struct lconv *lConv = localeconv();
	if (!pLocale)
		return pLocaleData;
	if (pLocaleData)
		pLocaleData = DestroyLocales(pLocaleData);
	pLocaleData = calloc(1, sizeof(LocaleDataType));
	if (pLocaleData) {
		pLocaleData->pLocaleCountry = CopyString(0, COUNTRY_ISO3166);
		pLocaleData->pLocaleLanguage = CopyString(0, LANGUAGE_ISO639);
		pLocaleData->pLocaleCurrency = CopyString(0, lConv->int_curr_symbol);
		pLocaleData->pLocaleShortDateTimePattern; // = GetLocaleShortDates();
		pLocaleData->pLocaleDecimalSeparator = CopyString(0, lConv->decimal_point);
		pLocaleData->pLocaleThousandSeparator = CopyString(0, lConv->thousands_sep);
		pLocaleData->pLocaleThousandGrouping = CopyString(0, GetLocaleThousandGrouping(lConv->grouping));
		switch (lConv->n_sign_posn) {
		case '0':
		case 0:
			pLocaleData->pLocaleNegativePrefix = CopyString(0, "(");
			pLocaleData->pLocaleNegativeSuffix = CopyString(0, ")");
			break;
		case '1':
		case 1:
			pLocaleData->pLocaleNegativePrefix = CopyString(0, lConv->negative_sign);
			pLocaleData->pLocaleNegativeSuffix = NULL;
			break;
		case '2':
		case 2:
			pLocaleData->pLocaleNegativePrefix = NULL;
			pLocaleData->pLocaleNegativeSuffix = CopyString(0, lConv->negative_sign);
			break;
		case '3':
		case 3:
			break;
		case '4':
		case 4:
			if (lConv->p_sign_posn) {
				pLocaleData->pLocaleNegativePrefix = CopyString(0, lConv->negative_sign);
				pLocaleData->pLocaleNegativeSuffix = NULL;
			}
			else {
				pLocaleData->pLocaleNegativePrefix = NULL;
				pLocaleData->pLocaleNegativeSuffix = CopyString(0, lConv->negative_sign);
			}
			break;
		}
		pLocaleData->pOperator = GetSystemUser(&pLocaleData->pOperator);
		{

			int i = 0;
			struct tm today;
			char tmpbuf[128];
			memset(&today, 0, sizeof(struct tm));
			today.tm_year = 3333 - 1900;
			today.tm_mon = 10;
			today.tm_mday = 22;
			strftime(tmpbuf, 128, "%x", &today);
			while (tmpbuf[i]) {
				switch (tmpbuf[i])
				{
				case '3':
					tmpbuf[i] = 'y';
					break;
				case '1':
					tmpbuf[i] = 'M';
					break;
				case '2':
					tmpbuf[i] = 'd';
					break;
				}
				i++;
			}
			pLocaleData->pLocaleShortDateTimePattern = (CopyString(0, tmpbuf));
		}
		{

			int i = 0;
			for (; i < 12; i++)
			{
				struct tm today;
				char tmpbuf[128];
				memset(&today, 0, sizeof(struct tm));
				today.tm_year = 3333 - 1900;
				today.tm_mon = i;
				today.tm_mday = 22;
				strftime(tmpbuf, 128, "%EB", &today);
				pLocaleData->pMonth[i] = (CopyString(0, tmpbuf));
			}
		}
		/*		{
			int n = 0;
			for (n = 0; n < NORMAL_MONTHS_PER_YEAR; n++)
				pLocaleData->pMonth[n] = GetLocaleMonth(n);
		}
	*/
	}
	return pLocaleData;
	//InstanceData.DefaultGeneralChartOfAccounts = GetVDUAccountingGeneralChartOfAccounts(CP_APP);
}

/********************************************************************

char *GetSystemUser(

*********************************************************************/
char *GetSystemUser(
	char **ppTarget)
{
	//GetUserName(pUserName, &dwUserNameSize);
	char *pTarget = CopyString(ppTarget, "System User Dummy");
	//	char *pUser = getenv("USERPROFILE");
	char *pUser = getenv("USERNAME");
	if (pUser && *pUser)
		pTarget = CopyString(ppTarget, pUser);
	return pTarget;
}

/********************************************************************

static char *GetLocaleThousandGrouping(

*********************************************************************/
static char *GetLocaleThousandGrouping(
	char *grouping)
{
	static char pLocaleThousandGrouping[2 * CHAR_MAX + 1];
	memset(pLocaleThousandGrouping, 0, 2 * CHAR_MAX + 1);
	int i = 0;
	for (i = 0; i < CHAR_MAX && grouping[i]; i++) {
		pLocaleThousandGrouping[(2 * i)] = grouping[i] + '0';
		pLocaleThousandGrouping[(2 * i) + 1] = ';';
		pLocaleThousandGrouping[(2 * i) + 2] = '0';
	}
	return pLocaleThousandGrouping;
}


