/*********************************************************************************

	RosettaFioProcess.h

	Alphabet Rosetta

	(c)Copyright 2017-2020  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

******************************************************************************************

The RosettaFioProcess.c file contains the basic OS independent logics for the application.

The RosettaFioProcess.h file contains the external declarations and definitions.

The basic logics is tight to the tab data and the interpretations sections
(at the end in this file), the rest of the RosettaFioProcess.c file is OS independent
application specific support functions.

The entrance is the Interpret File() to where a buffer of file content is
sent after being read. The file data is stored in a cloud of fragments of
data in shape of data structures, defined and handled in the RosettaFio.c and RosettaFio.h.
The RosettaFio.c and RosettaFio.h is machine generated code from XSD data specifications.

The LList.c handles the database and the Util.c app independent support functions.

The SearchSet-functions (in RosettaFioProcess.c add up new data in the linked list
database or updates old structures of data contents.

Files starting with OS identification contains OS specific UI functions. The
basic is very limited in App dependency while the app file is much app dependent.

The Tab-data represent a report (of various types) and the screen data of the
application is stored in a cell system in the RosettaView.c file. This data is
retried to fill the OS dependent screens of the applications. Only the last
part is OS dependent. Even tab data and menus are stored OS independently.

Many main tabs can be displayed (also be added or removed on the fly) and
each main tab point to its basic data and its display window data mirror.

The basic implementation is a Linked list data base solution in LList.c because
modern computers have some much memory that it is not an issue. It allows much
more reporting on the fly appearance of the application, like a word processor
file application. It gives the opportunity to make a version with more specific
data base systems like SQL is needed (for instance for real heavy huge accounting
use).

Content of RosettaFioProcess.c
1.) Basic program support features
2.) Locales
3.) Account coding
4.) Program texts
5.) Sum up Voucher data to Balances, month periods, quarter periods and section totals
6.) Search and get Data structures etc
7.) Sort functions
8.) Interpret, write, copy, merge and destroy the major data set

**********************************************************************************/
#if defined(__APPLE__)
#include "TargetConditionals.h"
#endif

#define ALL_BULLET						"\xc2\xA4"

// RosettaFioTypes.h
//#define FILE_DEFAULT					((BITMASK)0x0000)
//#define FILE_ACT_MASK					((BITMASK)0x1000)
//#define FILE_TYPE_MASK				((BITMASK)0x0f00)
//#define CONTENT_MASK					((BITMASK)0x00ff)
//#define FILE_ACT_OPEN					((BITMASK)0x0000)
//#define FILE_ACT_SAVE					((BITMASK)0x1000)
//#define FILE_ACT_IMPORT				((BITMASK)0x2000)
//#define FILE_TYPE_STD					((BITMASK)0x0000)
//#define FILE_TYPE_TXT					((BITMASK)0x0100)
//#define FILE_TYPE_CSV					((BITMASK)0x0200)
//#define FILE_TYPE_PDF					((BITMASK)0x0300)
//#define FILE_TYPE_TITLE				((BITMASK)0x0400)
//#define FILE_TYPE_PRINT				((BITMASK)0x0500)
#define FILE_TYPE_SIE					((BITMASK)0x0600)
#define FILE_TYPE_XGL					((BITMASK)0x0700)
#define CONTENT_NOTHING					((BITMASK)0x0000)
#define CONTENT_CHARTS					((BITMASK)0x0001)
#define CONTENT_BALANCE					((BITMASK)0x0002)
#define CONTENT_LEDGER					((BITMASK)0x0003)
#define CONTENT_VOUCHERS				((BITMASK)0x0004)
#define CONTENT_INVOICE					((BITMASK)0x0005)
#define CONTENT_BANKACS					((BITMASK)0x0006)
#define CONTENT_ERRORS					((BITMASK)0x000d)
#define CONTENT_ENTITY					((BITMASK)0x000e)
#define CONTENT_ANALYSE					((BITMASK)0x000f)
//#define CONTENT_MASK					((BITMASK)0x00ff)

#define CONTENT_LABEL_NOTHING			"Nothing"
#define CONTENT_LABEL_ANALYSE			"Analyse"
#define CONTENT_LABEL_PRESENTATION		"Presentation"
#define CONTENT_LABEL_ENTITY			"Entity"
#define CONTENT_LABEL_ERROR				"Error"
#define CONTENT_LABEL_OBJECTS			"Objects"
#define CONTENT_LABEL_ATTEST			"Attest"
#define CONTENT_LABEL_AUDIT				"Audit"
#define CONTENT_LABEL_ACCOUNTANT		"Accountant"

#define CONTENT_LABEL_CHARTS			"Charts"
#define CONTENT_LABEL_BALANCE			"Balances"
#define CONTENT_LABEL_LEDGER			"Ledger"
#define CONTENT_LABEL_VOUCHER			"Vouchers"
#define CONTENT_LABEL_INVOICE			"Invoice"
#define CONTENT_LABEL_BANKACS			"BankAccountStatement"

#define VOUCHERS_BY_SERIES				0
#define VOUCHERS_BY_DATE				1
#define VOUCHERS_BY_ORIGINAL			2

#define LINE_SEPARATOR					"----------------"

#define ACCOUNT_TYPE_SECTOR				"Sector"
#define ACCOUNT_TYPE_TOTAL				"Total"
#define ACCOUNT_TYPE_BALANCE			"Balance"
#define ACCOUNT_TYPE_RESULT				"Result"
#define ACCOUNT_TYPE_ACTIVE				"Active"
#define ACCOUNT_TYPE_PASSIVE			"Passive"
#define ACCOUNT_TYPE_PROFIT				"Profit"
#define ACCOUNT_TYPE_EARNING			"Earning"
#define ACCOUNT_TYPE_OUTCOME			"Outcome"
#define ACCOUNT_TYPE_TANGIBLE			"Tangible"
#define ACCOUNT_TYPE_STOCK				"Stock"
#define ACCOUNT_TYPE_RECEIVABLES		"Receivables"
#define ACCOUNT_TYPE_ASSETS				"Assets"
#define ACCOUNT_TYPE_EQUITIES			"Equities"
#define ACCOUNT_TYPE_PROVISIONS			"Provisions"
#define ACCOUNT_TYPE_LIABILITIES		"Liabilities"
#define ACCOUNT_TYPE_REVENUES			"Revenues"
#define ACCOUNT_TYPE_SUPPLIES			"Supplies"
#define ACCOUNT_TYPE_EXPENSES			"Expenses"
#define ACCOUNT_TYPE_STAFF				"Staff"
#define ACCOUNT_TYPE_LOSS_GAINS			"LossGains"
#define ACCOUNT_TYPE_FINANCIAL			"Financial"
#define ACCOUNT_TYPE_AUX				"Auxiliary"	

#define BAS2000							"BAS2000"	

#define DIM_TYPE_COSTCENTRE				"CostCentre"
#define DIM_TYPE_COSTBEARER				"CostBearer"
#define DIM_TYPE_POINTOFSALES			"PointOfSales"
#define DIM_TYPE_BRANCH					"Branch"
#define DIM_TYPE_PROJECT				"Project"
#define DIM_TYPE_EMPLYEE				"Employee"
#define DIM_TYPE_CUSTOMER				"Customer"
#define DIM_TYPE_SUPPLIER				"Supplier"
#define DIM_TYPE_INVOICE				"Invoice"

#define FILE_TYPE_LABEL_STD				"Rosetta"
#define FILE_TYPE_LABEL_SIE				"Sie"
#define FILE_TYPE_LABEL_XGL				"XBRL/GL"	
#define FILE_TYPE_LABEL_TXT				"TXT"	
#define FILE_TYPE_LABEL_CSV				"CSV"	
#define FILE_TYPE_LABEL_PDF				"PDF"	

#define FILE_TYPE_SUFFIX_TXT			".txt"
#define FILE_TYPE_SUFFIX_CSV			".csv"
#define FILE_TYPE_SUFFIX_STD			".rzt"
#define FILE_TYPE_SUFFIX_SIE_IE			".sie"
#define FILE_TYPE_SUFFIX_SIE_E			".se"
#define FILE_TYPE_SUFFIX_SIE_I			".si"
#define FILE_TYPE_SUFFIX_XGL			".xgl"
#define FILE_TYPE_SUFFIX_PDF			".pdf"

// en
#define TXT_ANALYSE						"Analyse"
#define TXT_ERRORS						"Errors"
#define TXT_CHART						"Charts"
#define TXT_OBJECTS						"Objects"
#define TXT_BALANCE						"Balances"
#define TXT_LEDGER						"Ledger"
#define TXT_VOUCHERS					"Vouchers"
#define TXT_INVOICE						"Invoice"
#define TXT_BANKACS						"Bank Account Statement"

#define TXT_NOTE						"Note"
#define TXT_FILE_CONTENT				"FileContent"
#define TXT_LAST_ENTRY_DATE				"LastEntryDate"
#define TXT_CURRENCY					"Currency"
#define TXT_MAIN_LANGUAGE				"MainLanguage"

#define TXT_ERROR_NO_FILE_NAME			"FailureNoFileName"
#define TXT_ERROR_NO_FILE_CONTENT		"FailureNoneIncorrectFileContent"
#define TXT_ERROR_EXPIRE				"FailureExpired"
#define TXT_ERROR_OS_VER				"FailureOsVersion"
#define TXT_ERROR_FILE_OPEN				"FailurePermission"
#define TXT_ERROR_FILE_CLOSE			"FailureCloseFile"
#define TXT_ERROR_MEMORY_MANAGEMENT		"FailureSystemError"

#define TXT_BUDGET						"Budget"
#define TXT_MONTHLY						"Monthly"
#define TXT_ANNUALLY					"Annually"
#define TXT_MAPPING						"Mapping"

#define TXT_MENU_OK						"Ok"
#define TXT_MENU_CANCEL					"Cancel"
#define TXT_MENU_FAILURE				"Failure"
#define TXT_MENU_QUANT					"Quant"
#define TXT_MENU_CHANGE					"Change"
#define TXT_MENU_SUM_MONTHS				"SumMonths"
#define TXT_MENU_FILE					"File"
#define TXT_MENU_EXIT					"Exit"
#define TXT_MENU_ABOUT					"About"
#define TXT_MENU_ABOUT_MESSAGE			"AboutMessage"
#define TXT_MENU_ABOUT_ROSETTA			"AboutRosetta"
#define TXT_MENU_ABOUT_CONTACT			"AboutContact"
#define TXT_MENU_EXPIRE					"Expire"
#define TXT_MENU_EDITION				"Edition"
#define TXT_MENU_RESPONSIBILITIES		"Responsibilities"
#define TXT_MENU_HELP					"Help"
#define TXT_MENU_HELP_MESSAGE			"HelpMessage"
#define TXT_MENU_HELP_DESKTOP			"HelpDesktop"
#define TXT_MENU_HELP_HANDSET			"HelpHandset"
#define TXT_MENU_SEARCH					"Search"
#define TXT_MENU_DRAGNDROP				"Dragdrop"
#define TXT_MENU_DIMOBJ					"Dimensions"
#define TXT_MENU_OPEN					"Open"
#define TXT_MENU_SAVE					"Save"
#define TXT_MENU_EDIT					"Edit"
#define TXT_MENU_COPY					"Copy"
#define TXT_MENU_REFRESH				"Refresh"
#define TXT_MENU_PRINT					"Print"
#define TXT_MENU_EMAIL					"eMail"
#define TXT_MENU_WASTE					"Waste"
#define TXT_MENU_SAVE_AS				"SaveAs"
#define TXT_MENU_SAVE_AS_ALL			"SaveAsAll"
#define TXT_MENU_TEXT					"Text"
#define TXT_MENU_SPREADSHEET			"Spreadsheet"
#define TXT_MENU_ENTITY					"Entity"
#define TXT_MENU_ACCOUNTS				"Accounts"
#define TXT_MENU_OBJECT					"Object"
#define TXT_MENU_SERIES					"Series"
#define TXT_MENU_DATE					"Date"
#define TXT_MENU_DEBIT					"Debit"
#define TXT_MENU_CREDIT					"Credit"
#define TXT_MENU_DO						"Do"
#define TXT_MENU_DO_NEW					"NewAccounting"
#define TXT_MENU_DO_YEAR				"NewYear"
#define TXT_MENU_DO_IMPORT				"Import"
#define TXT_MENU_NEW					"New"
#define TXT_MENU_EDIT					"Edit"
#define TXT_MENU_DELETE					"Delete"
#define TXT_MENU_EXAMINE				"Audit"
#define TXT_MENU_EXPAND					"Expand"
#define TXT_MENU_LINE_SELECT			"Select"
#define TXT_MENU_VIEW					"View"
#define TXT_MENU_VIEW_TAB_ADD			"TabAdd"
#define TXT_MENU_VIEW_TAB_DEL			"TabDelete"

#define TXT_MENU_LOGO					"Logotype"
#define TXT_MENU_URL					"Url"
#define TXT_MENU_WEBPAGE				"Webpage"
#define TXT_MENU_ADDRESS				"Address"
#define TXT_MENU_PHONE					"Phone"
#define TXT_MENU_SOCIALMEDIA			"Socialmedia"
#define TXT_MENU_CONTACTLOG				"Contactlog"
#define TXT_MENU_ALIAS					"Alias"
#define TXT_MENU_PKI					"Pki"
#define TXT_MENU_CRM					"Crm"
#define TXT_MENU_NOTE					"Note"
#define TXT_MENU_LEGAL_ID				"Legal id"
#define TXT_MENU_VAT_ID					"VAT id"
#define TXT_MENU_VAT_CURRENCY			"VAT currency"
#define TXT_MENU_JURISDICTION			"Entity Jurisdiction"
#define TXT_MENU_ENTITYTYPE				"Entity Type"
#define TXT_MENU_STATISTICAL			"Statistical class"
#define TXT_MENU_CASHACCOUNTING			"CashAccounting"
#define TXT_MENU_CONTACT				"Contact"

#define MENU_ERRORS						0x00 
#define MENU_EXIT						0x01 
#define MENU_HELP						0x02 
#define MENU_ABOUT						0x03 
#define MENU_SEARCH						0x04
#define MENU_SEARCH_BUTTON				0x05
#define MENU_DRAGNDROP					0x08
#define MENU_FILE						0x09 
#define MENU_REFRESH					0x0b
#define MENU_OPEN						0x0c 
#define MENU_IMPORT						0x0d 
#define MENU_SAVE						0x0e 
#define MENU_PRINT						0x0f 
#define MENU_SAVE_AS					0x10 
#define MENU_SAVE_AS_STD				0x11 
#define MENU_SAVE_AS_STD_ALL			0x12 
#define MENU_SAVE_AS_STD_CHART			0x13 
#define MENU_SAVE_AS_STD_BALANCE		0x14 
#define MENU_SAVE_AS_STD_VOUCHER		0x15 
#define MENU_SAVE_AS_STD_LEDGER			0x16 
#define MENU_SAVE_AS_STD_INVOICE		0x17 
#define MENU_SAVE_AS_STD_BANKACS		0x18 
#define MENU_SAVE_AS_STD_ERROR			0x19 
#define MENU_SAVE_AS_STD_ANALYSE		0x1a 
#define MENU_SAVE_AS_SIE				0x1b 
#define MENU_SAVE_AS_SIE_ALL			0x1c 
#define MENU_SAVE_AS_SIE_CHART			0x1d 
#define MENU_SAVE_AS_SIE_BALANCE		0x1e 
#define MENU_SAVE_AS_SIE_VOUCHER		0x1f 
#define MENU_SAVE_AS_XGL				0x20 
#define MENU_SAVE_AS_XGL_ALL			0x21 
#define MENU_SAVE_AS_XGL_CHART			0x22 
#define MENU_SAVE_AS_XGL_BALANCE		0x23 
#define MENU_SAVE_AS_XGL_VOUCHER		0x24 
#define MENU_TEXT						0x25 
#define MENU_SPREADSHEET				0x26
#define MENU_EDIT						0x2a
#define MENU_COPY						0x2b
#define MENU_EDIT_ENTITY				0x30 
#define MENU_EDIT_ACCOUNTS				0x31 
#define MENU_EDIT_OBJECTS				0x32 
#define MENU_EDIT_SERIES				0x33
#define MENU_EDIT_DATE					0x34
#define MENU_EDIT_DEBIT					0x35
#define MENU_EDIT_CREDIT				0x36
#define MENU_VIEW						0x37
#define MENU_VIEW_TAB_DEL				0x38
#define MENU_VIEW_TAB_ADD				0x39
#define MENU_VIEW_TAB_ADD_OBJECTS		0x3a
#define MENU_VIEW_TAB_ADD_BALANCE		0x3b
#define MENU_VIEW_TAB_ADD_LEDGER		0x3c
#define MENU_VIEW_TAB_ADD_VOUCHERS		0x3d
#define MENU_VIEW_TAB_ADD_INVOICE		0x3e
#define MENU_VIEW_TAB_ADD_BANKACC		0x3f
#define MENU_VIEW_TAB_ADD_ENTITY		0x40
#define MENU_VIEW_TAB_ADD_ERROR			0x41
#define MENU_EXAMINE					0x42 
#define MENU_EXPAND						0x43 
#define MENU_VIEW_ACCOUNTTYPE			0x50
#define MENU_VIEW_TOTAL					0x51
#define MENU_VIEW_BALANCE				0x52
#define MENU_VIEW_RESULT				0x53
#define MENU_VIEW_ACTIVE				0x54
#define MENU_VIEW_PASSIVE				0x55
#define MENU_VIEW_PROFIT				0x56
#define MENU_VIEW_EARNING				0x57
#define MENU_VIEW_OUTCOME				0x58
#define MENU_VIEW_TANGIBLE				0x59
#define MENU_VIEW_STOCK					0x5a
#define MENU_VIEW_RECEIVABLES			0x5b
#define MENU_VIEW_ASSETS				0x5c
#define MENU_VIEW_EQUITIES				0x5d
#define MENU_VIEW_PROVISIONS			0x5e
#define MENU_VIEW_LIABILITIES			0x5f
#define MENU_VIEW_REVENUES				0x60
#define MENU_VIEW_SUPPLIES				0x61
#define MENU_VIEW_EXPENSES				0x62
#define MENU_VIEW_STAFF					0x63
#define MENU_VIEW_LOSSGAINES			0x64
#define MENU_VIEW_FINANCIAL				0x65
#define MENU_VIEW_AUXILIARY_COSTS		0x66	

#define MENU_ALPHABET					0x6a
#define MENU_STORE_PLAY					0x6b
#define MENU_STORE_MS					0x6c
#define MENU_STORE_APPLE				0x6d

#define MENU_LOGO						0x80
#define MENU_URL						0x81
#define MENU_WEBPAGE					0x82
#define MENU_ADDRESS					0x83
#define MENU_EMAIL						0x84
#define MENU_PHONE						0x85
#define MENU_SOCIALMEDIA				0x86
#define MENU_CONTACTLOG					0x87
#define MENU_PKI						0x88
#define MENU_CRM						0x89
#define MENU_NOTE						0x8a
#define MENU_LEGAL_ID					0x8b
#define MENU_VAT_ID						0x8c
#define MENU_VAT_CURRENCY				0x8d
#define MENU_JURISDICTION				0x8e
#define MENU_ENTITYTYPE					0x8f
#define MENU_STATISTICAL				0x90
#define MENU_CASHACCOUNTING				0x91
#define MENU_CONTACT					0x92

#define MENU_DO							0xa0 
#define MENU_DO_NEW						0xa1
#define MENU_DO_YEAR					0xa2
#define MENU_DO_IMPORT					0xa3	   
#define MENU_DO_IMPORT_BALANCES			0xa4	   
#define MENU_DO_IMPORT_VOUCHERS			0xa5	   
#define MENU_HELP_MESSAGE				0xa6	   
#define MENU_ABOUT_MESSAGE				0xa7	   
#define MENU_DIMOBJ						0xae	   
#define MENU_BUDGET						0xaf	   
#define MENU_TXT_NOTE					0xb1
#define MENU_TXT_FILE_CONTENT			0xb2
#define MENU_TXT_LAST_ENTRY_DATE		0xb3
#define MENU_TXT_CURRENCY				0xb4
#define MENU_TXT_MAIN_LANGUAGE			0xb5
#define MENU_ERROR_NO_FILE_NAME			0xb6	   
#define MENU_ERROR_NO_FILE_CONTENT		0xb7	   
#define MENU_ERROR_EXPIRE				0xb8	   
#define MENU_ERROR_OS_VER				0xb9	   
#define MENU_ERROR_FILE_CLOSE			0xba	  
#define MENU_ERROR_FILE_OPEN			0xbb	   
#define MENU_ERROR_MEMORY_MANAGEMENT	0xbc	   
#define MENU_ADD_ITEM					0xc1	   
#define MENU_EDIT_ITEM					0xc2	   
#define MENU_DELETE_ITEM				0xc3
#define MENU_LINE_HOVER_TEXT			0xc5
#define MENU_LINE_SELECT				0xc6	   
#define MENU_LINE_OPEN					0xc7	   
#define MENU_QUANT						0xca   
#define MENU_CHANGE						0xcb	   
#define MENU_SUM_MONTHS					0xcc	   
#define MENU_FAILURE					0xcd	   
#define MENU_CANCEL						0xce	   
#define MENU_OK							0xcf	   
#define MENU_SERIES						0xd0	   
#define MENU_SERIES_FIRST				0xd1	   
#define MENU_SERIES_LAST				0xff	   

#define MENU_PERIOD_DRAWER				0x100000
#define MENU_OBJECT_DRAWER				0x200000


#define EMOJI_MENU_EXIT					"\xF0\x9F\x9A\xAA"
#define EMOJI_MENU_FILE					"\xF0\x9f\x93\x82"
#define EMOJI_COPY_EDIT					""
#define EMOJI_MENU_EDIT					"\xE2\x9c\x91"
#define EMOJI_MENU_ABOUT				"\xE2\x89\x88"
#define EMOJI_MENU_HELP					"\xE2\x9D\x94"
#define EMOJI_MENU_SEARCH				"\xF0\x9F\x94\x8D"
#define EMOJI_MENU_REFRESH				"\xF0\x9F\x94\x84"
#define EMOJI_MENU_OPEN					"\xF0\x9f\x93\x82"
//#define EMOJI_MENU_IMPORT				"\xE2\x9E\xA1"
#define EMOJI_MENU_IMPORT				"\xE2\x87\x92"
#define EMOJI_MENU_SAVE_SPREADSHEET		"\xF0\x9F\x93\x9C"
#define EMOJI_MENU_SAVE_TEXT			"\xF0\x9F\x93\x83"
#define EMOJI_MENU_SAVE					"\xF0\x9f\x92\xbe"
#define EMOJI_MENU_SAVE_AS				"\xF0\x9f\x93\x81"
#define EMOJI_MENU_PRINT				"\xE2\x8E\x99"
#define EMOJI_MENU_EMAIL				"\xE2\x9C\x89"
#define EMOJI_MENU_WASTE				"\xF0\x9f\x97\x91"
#define EMOJI_MENU_DELETE				"\xE2\x9D\x8C"
#define EMOJI_MENU_DO					"\xF0\x9F\x94\xA7"
#define EMOJI_MENU_ADD					"\xE2\x9c\x9b"
#define EMOJI_MENU_NEW					"\xE2\x9c\x9a"
#define EMOJI_MENU_NEW2					"\xE2\x9c\x9c"
#define EMOJI_MENU_VIEW					"\xF0\x9F\x91\x93"

#define EMOJI_MENU_LOGO					"\xF0\x9F\x93\x82"
#define EMOJI_MENU_URL					"\xF0\x9F\x93\x82"
#define EMOJI_MENU_WEBPAGE				"\xF0\x9F\x93\x82"
#define EMOJI_MENU_ADDRESS				"\xF0\x9F\x93\xAB"
#define EMOJI_MENU_PHONE				"\xE2\x98\x8E"
#define EMOJI_MENU_SOCIALMEDIA			"\xE2\x98\xBA"
#define EMOJI_MENU_CONTACTLOG			""
#define EMOJI_MENU_PKI					""
#define EMOJI_MENU_CRM					""
#define EMOJI_MENU_NOTE					"\xF0\x9F\x93\x9D"
#define EMOJI_MENU_LEGAL_ID				""
#define EMOJI_MENU_VAT_ID				""
#define EMOJI_MENU_VAT_CURRENCY			""
#define EMOJI_MENU_JURISDICTION			""
#define EMOJI_MENU_ENTITYTYPE			""
#define EMOJI_MENU_STATISTICAL			""
#define EMOJI_MENU_CASHACCOUNTING		""
#define EMOJI_MENU_CONTACT				""

#define MVA_LEFT						(BITMASK)0x0000 // Same as HDF_LEFT LVCFMT_LEFT
#define MVA_RIGHT						(BITMASK)0x0001 // Same as HDF_RIGHT LVCFMT_RIGHT
#define MVA_CENTER						(BITMASK)0x0002 // Same as HDF_CENTER LVCFMT_CENTER
#define MVA_JUSTIFYMASK					(BITMASK)0x0003 // Same as HDF_JUSTIFYMASK LVCFMT_JUSTIFYMASK
#define MVA_DYNAMIC_TEXT				(BITMASK)0x0004
#define MVA_RIGHT_COLUMN				(BITMASK)0x0008
#define MVA_DYNAMIC_MASK				(BITMASK)0x000c

#define AMOUNT_WIDTH					64
#define QUANT_WIDTH						57
#define THOUSANDSET_WIDTH				40
#define DATE_WIDTH						98
#define ACOCUNT_WIDTH					50
#define ACOCUNT_NAME_WIDTH				250
#define VOUCHER_TITLE_WIDTH				250
#define ID_WIDTH						68
#define DATE_WIDTH_BOLD					(DATE_WIDTH + 9)

// Basic program support features
char *ProgramTextsLanguageSupport(char *pLang);
char *GetSystemProgram(char **ppTarget);
char *GetSystemProgramVersion(char **ppTarget);
char *GetWorkDataName(void *vpWorkData);
char *GetContent(void *vpAccounting);
char *AppGetContentFromBm(BITMASK bmCommand);
BITMASK AppGetBmFromContent(char *pContent);
BITMASK GetContentParameters(void *vpWorkData, int iFiletype);
char *GetWorkDataLogotype(void *vpWorkData);
char *GetWorkDataCurrency(void *vpWorkData);
DATEINT GetLastEntryDate(AccountingType *pAccounting);
DATEINT GetLedgerLastEntryDate(void *vpAccounting);
char *GetWorkDataRange(void *vpWorkData);

#if defined(ANDROID) || defined(IOS)
// Aux functions
int GetDimTotals(void *vpWorkData, BOOL bAudit);
COUNTER *GetDimIds(void *vpWorkData, BOOL bAudit);
char *GetDimTitle(void *vpWorkData, char *pLanguage, COUNTER cDimId);
char **GetObjIds(void *vpWorkData, BOOL bAudit, COUNTER cDimId);
char *GetObjTitle(void *vpWorkData, char *pLanguage, COUNTER cDimId, char *pObjId);
int GetObjTotals(void *vpWorkData, BOOL bAudit, COUNTER cDimId);
#endif
// Locales
LocaleDataType *DestroyLocales(LocaleDataType *pLocaleData);

// Account coding
//static COUNTER CounterAccountCoding(char *pAccountType);
AccountingType *ConvertAccountCoding(AccountingType *pSourceAccounting, BITMASK bmFileType);
AccountingType *ExpandSieAccountTypes(AccountingType *pSourceAccounting, BITMASK bmFileType);

// Program texts
char *GetEnumApplicationTextType(char *pType, char *pLang);
char *GetEnumDepreciationPeriodsType(char *pType, char *pLang);
char *GetEnumFolderTypeText(char *pType, char *pLang);
char *GetEnumAccountTitleTypeText(char *pType, char *pLang);
char *AppMajorTexts(int iTextItem, char *pLocaleLanguage, BOOL bEmoji);
char *AppMenuTexts(int iTextItem, LocaleDataType *pLocaleData, BOOL bEmoji, BOOL bHwKeyb);
char *CreateGeneralTitle(InstanceDataType *pInstanceData, char *pFilePathName, BITMASK bmCommand, TabDataType *pTabData, char *pCurrency, DATEINT LastDate, size_t ztMaxPathLength);
char *NuberedTabTexts(char *pTabText, NUMERIC nIndex);

// Sum up Voucher data to Balances, month periods, quarter periods and section totals
void SumUpAccounting(AccountingType *pAccounting);
void SumUpBalanceTotals(AccountingType *pAccounting);

// Search and get Data structures etc
LLIST AddTextListItem(void *pTopStruct, void *pParentStruct, LLIST pListText, char *pString, char *plang);
char *GetTextListText(LLIST pListText, char *pLanguage);
TextType *AssembleText(void *pTopStruct, void *pParentStruct, char *pTextString, char *pLang);
SeriesIdType *AseembleSeriesId(void *pTopStruct, void *pParentStruct, char *pSeries, COUNTER pId);
OperatorInfoType *AseembleOperatorInfo(void *pTopStruct, void *pParentStruct, char *pOperatorInfoString, DATEINT Date);
char *GetDeprecatedText(OperatorInfoType *pDeprecatedInfo, char *pTitle, char *pDeprecatedText, size_t DeprecatedTextLength);
AccountCodeType *AseembleAccountCode(void *pTopStruct, void *pParentStruct, char *pCodeType, char *pIssuer, char *pApplicableDomain, char *pValidYear, char *pCode, char *pDescription, char *pDescriptionLang);
TextType *SearchSetText(LLIST pListText, TextType *pSearchText);
char *SearchSetToken(LLIST pListToken, char *pSearchToken);
void *GetEntity(void *vpAccounting);
//static ContactType *GetContact(void *vpParty, char *pName);
ContactType *AddContact(void *vpParty, char *pName);
AccountType *GetAccount(void *vpAccounting, char *pSearchAccountId);
AccountType *SearchSetAccount(LLIST pListAccount, AccountType *pSearchAccount);
AccountType *AddAccount(void *vpAccounting, char *pAccountId);
AccountCodeType *SearchSetAccountCode(LLIST pListAccountCode, AccountCodeType *pSearchAccountCode);
LLIST GetObjectsChart(InstanceDataType *pInstanceData);
DimensionType *GetDimension(LLIST pListDimensions, COUNTER cDimensionId);
DimensionType *SearchSetDimension(LLIST pListDimension, DimensionType *pSearchDimension);
DimensionType *AddDimension(void *vpAccounting, COUNTER cDimensionId);
void *GetDimObjectAccounting(void *vpWorkData, COUNTER cDimensionId, char *pObjectId);
ObjectType *GetDimObject(LLIST pListDimension, COUNTER cDimensionId, char *pObjectId);
ObjectType *SearchSetObject(LLIST pListDimension, ObjectType *pSearchObject);
ObjectType *AddDimObject(void *vpAccounting, COUNTER cDimensionId, char *pObjectId);
void *GetLedger(void *vpAccounting, NUMERIC nLedgerId);
LedgerType *AddLedger(void *vpAccounting, NUMERIC nLedgerId, DATEINT FiscalYearStart, DATEINT FiscalYearEnd);
LedgerType *GetLedgerByDate(void *vpAccounting, DATEINT Date);
LedgerType *GetLedgerDynamic(void *vpAccounting, DATEINT Date);
LedgerType *SearchSetLedger(LLIST pListLedger, LedgerType *pSearchLedger);
void *GetBalanceReport(void *vpLedger, void *vpObject);
BalanceReportType *SearchSetBalanceReport(LLIST pBalanceReportList, BalanceReportType *pSearchBalanceReport);
void *GetBalanceAccount(void *vpBalanceReport, void *vpAccount);
BalanceAccountType *SearchSetBalanceAccount(LLIST pListBalanceAccount, BalanceAccountType *pSearchBalanceAccount);
void *GetBalanceAccountPeriod(BalanceAccountType *pBalanceAccount, DATEINT PeriodEnd, char * pSelTabSubText);
BalanceAccountPeriodType *SearchSetBalanceAccountPeriodDynamic(BalanceAccountType *pBalanceAccount, BalanceAccountPeriodType *pSearchBalanceAccountPeriod, char *pSelTabSubText);
BalanceAccountPeriodType *SearchSetBalanceAccountPeriod(LLIST pListBalanceAccountPeriod, BalanceAccountPeriodType *pSearchBalanceAccountPeriod);
BalanceTotalPeriodType* GetBalanceTotalPeriod(LLIST pListBalanceTotalPeriod, DATEINT End);
VoucherType *SearchSetVoucher(LLIST pListVoucher, VoucherType *pSearchVoucher);
void CheckEnum(char **ppEnumData, EnumType *pEnumList, char *pEnumRead);

// Sort functions
LLIST GetSortedDataStructList(LLIST pListDataStructs, int SortOption);
void Sort(AccountingType *pAccounting, int VoucherSortOption);
void RefreshAccountingData(void *vpAccounting);

// Interpret, write, copy, merge and destroy the major data set
BOOL IsData(AccountingType *pAccounting, ObjectType *pObject, DATEINT PeriodDate, char *PeriodText, char *pSeries);
BOOL IsPeriodic(void *vpAccounting);
BOOL IsLedgerPeriodic(LedgerType *pLedger);
BOOL IsBudget(void *vpAccounting, AccountType *pAccount, char *pAccountType, ObjectType *pObject, DATEINT PeriodDate, char *pPeriodText);
BOOL IsVouchers(void *vpAccounting);
BOOL IsDimensionActive(void *vpAccounting, DimensionType *pDimension);
BOOL IsObject(void *vpAccounting);
void CheckActive(void *vpAccounting, void *vpObject, DATEINT dPeriod, char *pPeriodTabText, BOOL bVoucherOnly, char *pVoucherSeries);
InstanceDataType *DestroyInstanceData(InstanceDataType *pInstanceData);
LLIST GetSeriesList(void *vpAccounting, TabDataType *pTabData, NUMERIC nLedgerYearId);
size_t MergeAccountingInstanceData(size_t ztOldInstanceData, size_t ztNewInstanceData);
int ExternalWriteWorkData(InstanceDataType *pInstanceData, BITMASK bmCommand, char *pFilePathName);
size_t InterpretFile(char *pBuffer, char *pFilePathName, LocaleDataType *pLocaleData);
