/*********************************************************************************

	RosettaFioSie.c

	Alphabet Rosetta

	(c)Copyright 2017-2020  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

******************************************************************************************

The RosettaFioSie.c file contains functions to interpret and write SIE-files (www.sie.se)
into the Alphabet Rosetta Accounting basic data logics and database system.

The RosettaFioSie.h file contains the external declarations and definitions.

Content of RosettaFioSie.c
1.) Interpret SIE file data
2.) Create SIE file data

**********************************************************************************/

// Interpret SIE file data
void *OpenSieFile(char*pFilePathName, LLIST plErrorList);
void *InterpretSie(char *pFilePathName, char *pBuffer, LLIST plErrorList);

// Create SIE file data
int FileWriteSie(
	char *pFilePathName,
	void *vpAccounting,
	char *pContent,
	char *pGenUserName,
	BOOL CompactFileOption,
	BOOL LeanChartsOption,
	BOOL QuantityOption,
	BOOL PeriodOption,
	BOOL ObjectOption,
	BOOL AccountSortOption,
	BOOL DimensionSortOption,
	BOOL ObjectSortOption,
	int  VoucherSortOption,
	char *pInitialDate,
	char *pFinalDate,
	char *FilterEnteredBy,
	char *pProgramName,
	char *pProgramVersion,
	char *pLanguage,
	BOOL bUtf8);

int StreamWriteSie(
	FILE *pStream,
	void *vpAccounting,
	char *pContent,
	char *pGenUserName,
	BOOL CompactFileOption,
	BOOL LeanChartsOption,
	BOOL QuantityOption,
	BOOL PeriodOption,
	BOOL ObjectOption,
	BOOL AccountSortOption,
	BOOL DimensionSortOption,
	BOOL ObjectSortOption,
	int  VoucherSortOption,
	char *pInitialDate,
	char *pFinalDate,
	char *FilterEnteredBy,
	char *pProgramName,
	char *pProgramVersion,
	char *pLanguage);
