/*********************************************************************************

	RosettaFioUtils.c

	Alphabet Rosetta

	(c)Copyright 2017-2020  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

******************************************************************************************

The RosettaFioUtils.c file contains the basic OS and app independent support functions.

The RosettaFioUtils.h file contains the external declarations and definitions.

Content of RosettaFioUtils.c
1.) Date and time functions
2.) String to and to String functions
3.) String UTF8 cmp functions
4.) String functions
5.) File functions
6.) Xml Syntax functions
7.) Link LIst functions
**********************************************************************************/
#include "RosettaFioTypes.h"
#include <ctype.h> 
#include <time.h>
#include <stdlib.h>
/********************************************************************

Date and time functions

*********************************************************************/
/********************************************************************

char *GetSystemYear(void)

********************************************************************/
char *GetSystemYear(void)
{
	static char DateString[128];

	time_t ltime;
	//	_tzset();
	time(&ltime);
	strftime(DateString, 128,
		"%Y", localtime(&ltime));
	return DateString;
}

/********************************************************************

DATEINT GetSystemDate(void)

********************************************************************/
DATEINT GetSystemDate(void)
{
	char DateString[128];
	time_t ltime;
	//	_tzset();
	time(&ltime);
	strftime(DateString, 128,
		"%Y%m%d", localtime(&ltime));
	return StringToDate(DateString);
}

/********************************************************************

DATETIMEINT GetSystemDateTime(void)

********************************************************************/
DATETIMEINT GetSystemDateTime(void)
{
	char DateString[128];
	time_t ltime;
	//	_tzset();
	time(&ltime);
	strftime(DateString, 128,
		"%Y%m%d%H%M%S", localtime(&ltime));
	return StringToDateTime(DateString);
}

/********************************************************************

BOOL IsLeapYr(int iYear)

********************************************************************/
//  IsLeapYr determines if a given year value is a leap year. The
//  function returns TRUE if the current year is a leap year, and 
//  FALSE otherwise.
BOOL IsLeapYr(int iYear)
{
	BOOL fLeapYr = FALSE;

	// If the year is evenly divisible by 4 and not by 100 but year by 400, then this 
	// is a leap year.
	if((!(iYear % 4))				// Each even fourth year
		&& ((iYear % 100)			// Not each even 100 year
		|| (!(iYear % 400))))	// But each even 400 year
		fLeapYr = TRUE;

	return (fLeapYr);
}

/********************************************************************

DATEINT GetFirstDateInYear(DATEINT Date)

*********************************************************************/
DATEINT GetFirstDateInYear(DATEINT Date)
{
	return ((Date / 10000) * 10000) + 101;
}


/********************************************************************

DATEINT GetNextMonthLastDate(DATEINT Date)

*********************************************************************/
DATEINT GetNextMonthLastDate(DATEINT Date)
{
	DATEINT Year = Date / 10000;
	DATEINT Month = (Date / 100) % 100;
	DATEINT Day = Date % 100;

    DATEINT iDaysInMonths[12] = {31,28,31,30,31,30,31,31,30,31,30,31};

	Month++;
	if(Month>50) // Subtraction overflow
		Month = 12 - (100 - Month);
	else if(Month>12)  // Addition overflow
		{
		Month--;
		// Swap temp to zero base
		Year += (Month / 12);
		Month = (Month % 12);
		Month++;
		}
	if(IsLeapYr((int)Year))
		iDaysInMonths[2 - 1] = 29;
	Day = iDaysInMonths[Month - 1];

	return (Year * 10000) + (Month * 100) + Day;
}

/********************************************************************

DATEINT GetPrevMonthLastDate(DATEINT Date)

*********************************************************************/
DATEINT GetPrevMonthLastDate(DATEINT Date)
{
	DATEINT Year = Date / 10000;
	DATEINT Month = (Date / 100) % 100;
	DATEINT Day = Date % 100;

	int iDaysInMonths[12] = {31,28,31,30,31,30,31,31,30,31,30,31};

	if(!Month)
	{
		Month = 12;
		Year--;
	}
	else if (Month > 50) // Subtraction overflow
		Month = 12 - (100 - Month);
	else if(Month>12)
	{
		Month--; // 0-base
		Year += (Month / 12);
		Month %= 12;
		Month++;
	}
	Month--;
	if(Month<1)
	{
		Year--;
		Month = 12;
	}
	if(IsLeapYr((int)Year))
		iDaysInMonths[2 - 1] = 29;
	Day = iDaysInMonths[Month - 1];

	return (Year * 10000) + (Month * 100) + Day;
}

/********************************************************************

DATEINT GetNextDate(DATEINT Date)

*********************************************************************/
DATEINT GetNextDate(DATEINT Date)
{
	if(Date)
	{
		DATEINT Year = Date / 10000;
		DATEINT Month = (Date / 100) % 100;
		DATEINT Day = Date % 100;

		int iDaysInMonths[13] = {31,28,31,30,31,30,31,31,30,31,30,31,0};

		if(!Month)
		{
			Month = 12;
			Year--;
		}
		else if (Month > 50) // Subtraction overflow
			Month = 12 - (100 - Month);
		else if(Month>12)
		{
			Month--; // 0-base
			Year += (Month / 12);
			Month %= 12;
			Month++;
		}
		Day++;
		if(IsLeapYr((int)Year))
			iDaysInMonths[2 - 1] = 29;
		if(Day > iDaysInMonths[Month - 1])
		{
			Day -= iDaysInMonths[Month - 1];
			Month++;
			if(Month > 12)
			{
				Month--;
				// Swap temp to zero base
				Year += (Month / 12);
				Month = (Month % 12);
				Month++;
			}
		}
		return (Year * 10000) + (Month * 100) + Day;
	}
	return 0;
}


/********************************************************************

DATEINT GetPrevDate(DATEINT Date)

*********************************************************************/
DATEINT GetPrevDate(DATEINT Date)
{
	if(Date)
	{
		DATEINT Year = Date / 10000;
		DATEINT Month = (Date / 100) % 100;
		DATEINT Day = Date % 100;

		int iDaysInMonths[12] = {31,28,31,30,31,30,31,31,30,31,30,31};

		Day--;
		if(!Day)
		{
			Month--;
			if(!Month)
			{
				Month = 12;
				Year--;
				Day = 31;
			}
			else if (Month > 50) // Subtraction overflow
				Month = 12 - (100 - Month);
			else if(Month>12)
			{
				Month--; // 0-base
				Year += (Month / 12);
				Month %= 12;
				Month++;
				if(IsLeapYr((int)Year))
					iDaysInMonths[2 - 1] = 29;
				Day = iDaysInMonths[Month - 1];
			}
			else
			{
				if(IsLeapYr((int)Year))
					iDaysInMonths[2 - 1] = 29;
				Day = iDaysInMonths[Month - 1];
			}
		}
		return (Year * 10000) + (Month * 100) + Day;
	}
	return 0;
}


/********************************************************************

DATEINT GetPeriodDate(DATEINT BalancePeriodDate, char *pPeriodText)

*********************************************************************/
DATEINT GetPeriodDate(DATEINT BalancePeriodDate, char *pPeriodText) 
{
	if(pPeriodText && BalancePeriodDate)
	{
		DATEINT PeriodDate = GetLastDateInMonth(BalancePeriodDate);
		if(!strncmp(pPeriodText, "A0", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231 - 10000;
		else if(!strncmp(pPeriodText, "A1", 2))
				PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231;
		else if(!strncmp(pPeriodText, "S0", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231 - 10000;
		else if(!strncmp(pPeriodText, "S2", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231;
		else if(!strncmp(pPeriodText, "S1", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 630;
		else if(!strncmp(pPeriodText, "T0", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231 - 10000;
		else if(!strncmp(pPeriodText, "T3", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231;
		else if(!strncmp(pPeriodText, "T2", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 831;
		else if(!strncmp(pPeriodText, "T1", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 430;
		else if(!strncmp(pPeriodText, "Q4", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231;
		else if(!strncmp(pPeriodText, "Q0", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 1231 - 10000;
		else if(!strncmp(pPeriodText, "Q3", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 930;
		else if(!strncmp(pPeriodText, "Q2", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 630;
		else if(!strncmp(pPeriodText, "Q1", 2))
			PeriodDate = ((BalancePeriodDate / 10000) * 10000) + 331;
		//		while (GetLastDateInPeriod(BalancePeriodDate, *pPeriodText) < PeriodDate)
		//			PeriodDate -= 10000;
		return PeriodDate;
	}
	return 0;
}

/********************************************************************

DATEINT GetLastDateInPeriod(DATEINT Date, char *pPeriodLabel)

*********************************************************************/
DATEINT GetLastDateInPeriod(DATEINT Date, char *pPeriodLabel)
{
	DATEINT Year = Date / 10000;
	DATEINT Month = (Date / 100) % 100;
	//DATEINT Day = Date % 100;
	char chPeriodType = MONTH;  
	if(pPeriodLabel && ((pPeriodLabel[1] >= '0') && (pPeriodLabel[1] <= '9')))
		chPeriodType = *pPeriodLabel;
	switch(chPeriodType)
	{
	case QUARTER:
		if(Month % 3)
		{
			Month /= 3;
			Month++;
			Month *= 3;
		}
		return GetLastDateInMonth((Year * 10000) + (Month * 100) + 1);
		break;
	case TERTIAL:
		if(Month % 4)
		{
			Month /= 4;
			Month++;
			Month *= 4;
		}
		return GetLastDateInMonth((Year * 10000) + (Month * 100) + 1);
		break;
	case SEMI:
		if(Month % 6)
		{
			Month /= 6;
			Month++;
			Month *= 6;
		}
		return GetLastDateInMonth((Year * 10000) + (Month * 100) + 1);
		break;
	case ANNUAL:
		if(Month % 12)
		{
			Month /= 12;
			Month++;
			Month *= 12;
		}
		return GetLastDateInMonth((Year * 10000) + (Month * 100) + 1);
		break;
	case MONTH:
	default:
		return GetLastDateInMonth(Date);
		break;
	}

}


/********************************************************************

DATEINT GetLastDateInMonth(DATEINT Date)

*********************************************************************/
DATEINT GetLastDateInMonth(DATEINT Date)
{
	if (Date)
	{
		DATEINT Year = Date / 10000;
		DATEINT Month = (Date / 100) % 100;
		//DATEINT Day = Date % 100;

		char iDaysInMonths[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

		if (!Month)
		{
			Month = 12;
			Year--;
		}
		else if (Month > 50) // Subtraction overflow
			Month = 12 - (100 - Month);
		else if (Month > 12)
		{
			// Nuts
			Month--;
			// Swap temp to zero base
			Year += (Month / 12);
			Month = (Month % 12);
			Month++;
		}
		if (IsLeapYr((int)Year))
			iDaysInMonths[2 - 1] = 29;
		return (Year * 10000) + (Month * 100) + iDaysInMonths[Month - 1];
	}
	return 0;
}

/********************************************************************

BOOL CheckDateTime(DATEINT Date, TIMEINT Time)
*
*********************************************************************/
BOOL CheckDateTime(DATEINT Date, TIMEINT Time)
{
	if(Date)
	{
		if(Date<17530301)
			return FALSE; // Invalid year (system cant take previous years)
		if(Date>99991231)
			return FALSE; // Invalid year (system cant take previous years)
		if((((Date / 100) % 100)>12) || (((Date / 100) % 100)<1))
			return FALSE; // Invalid month
		if((((((Date / 100) % 100) == 1)
			|| (((Date / 100) % 100) == 3)
			|| (((Date / 100) % 100) == 5)
			|| (((Date / 100) % 100) == 7)
			|| (((Date / 100) % 100) == 8)
			|| (((Date / 100) % 100) == 10)
			|| (((Date / 100) % 100) == 12))
			&& (((Date) % 100)>31)) || (((Date) % 100)<1))
			return FALSE; // Invalid day
		else if(((Date / 100) % 100) == 2)
		{
			extern BOOL IsLeapYr(int iYear);
			if((Date % 100)<1)
				return FALSE; // Invalid day
			else if(IsLeapYr((int)(Date / 10000)))
			{
				if(((Date) % 100)>29)
					return FALSE; // Invalid day
			}
			else if(((Date) % 100)>28)
				return FALSE; // Invalid day
		}
		else if((((((Date / 100) % 100) == 4)
			|| (((Date / 100) % 100) == 6)
			|| (((Date / 100) % 100) == 9)
			|| (((Date / 100) % 100) == 11))
			&& (((Date) % 100)>30)) || (((Date) % 100)<1))
			return FALSE; // Invalid day
	}

	if(Time)
	{
		if((Time / 10000)>23)
			return FALSE; // Invalid hour
		if(((Time / 100) % 100)>59)
			return FALSE; // Invalid minute
		if(((Time) % 100)>59)
			return FALSE; // Invalid second
	}
	return TRUE;
}

/*************************************************************

DATEINT AddDays(DATEINT BaseDate, int iDays)

*************************************************************/
DATEINT AddDays(DATEINT BaseDate, int iDays)
{
	while(iDays)
	{
		BaseDate = GetNextDate(BaseDate);
		iDays--;
	}
	return BaseDate;
}

/*************************************************************

DATEINT AddMonths(DATEINT BaseDate, int iMonths)

*************************************************************/
DATEINT AddMonths(DATEINT BaseDate, int iMonths)
{
	DATEINT NewDate = BaseDate + (100 * ((DATEINT)iMonths));
	int iAddYears = (int)(((NewDate / 100) % 100) - 1 / 12);
	if(iAddYears)
	{
		NewDate -= (1200 * ((DATEINT)iAddYears));
		NewDate += (10000 * ((DATEINT)iAddYears));
	}
	if(NewDate > GetLastDateInMonth(NewDate))
		NewDate = GetLastDateInMonth(NewDate);
	return NewDate;
}

/*************************************************************

DATEINT AddYears(DATEINT BaseDate, int iYears)

*************************************************************/
DATEINT AddYears(DATEINT BaseDate, int iYears)
{
	return BaseDate += (10000 * ((DATEINT)iYears));
}

/********************************************************************

String to and to String functions

*********************************************************************/
/********************************************************************

BOOL DecimalCmp(QUANT Decimal1, QUANT Decimal2)

*********************************************************************/
long Strtol(char const *pString);
BOOL DecimalCmp(QUANT Decimal1, QUANT Decimal2)
{
#define EPSILON 0.0000001   // Define your own tolerance
#define FLOAT_EQ(x,v) (((v - EPSILON) < x) && (x <( v + EPSILON)))

	if(FLOAT_EQ(Decimal1, Decimal2))
		return TRUE;
	return FALSE;
}

/********************************************************************

int Strtoi(char const* String)

*********************************************************************/
int Strtoi(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return (int)strtol(pString, (char **)NULL, 10);
}

/********************************************************************

long Strtol(char const *pString)

*********************************************************************/
long Strtol(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return strtol(pString, (char **)NULL, 10);
}

/********************************************************************

long long Strtoll(char const *pString)

*********************************************************************/
long long Strtoll(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return strtoll(pString, (char **)NULL, 10);
}

/********************************************************************

unsigned int Strtoui(char const *pString)

*********************************************************************/
unsigned int Strtoui(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return (unsigned int)strtoul(pString, (char **)NULL, 10);
}

/********************************************************************

unsigned long Strtoul(char const *pString)

*********************************************************************/
unsigned long Strtoul(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return strtoul(pString, (char **)NULL, 10);
}

/********************************************************************

unsigned long long Strtoull(char const *pString)

*********************************************************************/
unsigned long long Strtoull(char const *pString)
{
	if ((!pString) || (!*pString))
		return 0;
	return strtoull(pString, (char **)NULL, 10);
}

/********************************************************************

char *itoStr(int Value, char *pString)

*********************************************************************/
char *itoStr(int Value, char *pString)
{
	sprintf(pString, "%d", Value);
	return pString;
}

/********************************************************************

char *ltoStr(long Value, char *pString)

*********************************************************************/
char *ltoStr(long Value, char *pString)
{
	sprintf(pString, "%ld", Value);
	return pString;
}

/********************************************************************

char *lltoStr(long long Value, char *pString)

*********************************************************************/
char *lltoStr(long long Value, char *pString)
{
	sprintf(pString, "%lld", Value);
	return pString;
}

/********************************************************************

char *utoStr(unsigned int Value, char *pString, int Radix)

*********************************************************************/
char *utoStr(unsigned int Value, char *pString, int Radix)
{
	switch (Radix)
	{
	case 16:
		sprintf(pString, "%x", Value);
	case 8:
		sprintf(pString, "%o", Value);
	default:
		sprintf(pString, "%u", Value);
	}
	return pString;
}

/********************************************************************

char *ultoStr(unsigned long Value, char *pString, int Radix)

*********************************************************************/
char *ultoStr(unsigned long Value, char *pString, int Radix)
{
	switch (Radix)
	{
	case 16:
		sprintf(pString, "%lx", Value);
	case 8:
		sprintf(pString, "%lo", Value);
	default:
		sprintf(pString, "%lu", Value);
	}
	return pString;
}

/********************************************************************

char *ulltoStr(unsigned long long Value, char *pString, int Radix)

*********************************************************************/
char *ulltoStr(unsigned long long Value, char *pString, int Radix)
{
	switch (Radix)
	{
	case 16:
		sprintf(pString, "%llx", Value);
	case 8:
		sprintf(pString, "%llo", Value);
	default:
		sprintf(pString, "%llu", Value);
	}
	return pString;
}

/********************************************************************

BOOL StringToBool(const UCHAR *pString)

*********************************************************************/
BOOL StringToBool(const char *pString)
{
	if (pString && *pString)
		{
		StrLwr((UCHAR *)pString);
		if (!strcmp(pString, "false"))
			return 0;
		else if (!strcmp(pString, "true"))
			return 1;
		else if (!strcmp(pString, "0"))
			return 0;
		else if (!strcmp(pString, "1"))
			return 1;
		else
			return 0;
		}
	else
		return 0;
}

/********************************************************************

char *BoolToStr(BOOL Bool, char *pString)

*********************************************************************/
char *BoolToStr(BOOL Bool, char *pString)
{
	switch (Bool)
	{
	case 0:
		return "false";
	default:
		return "true";
	}
}

/********************************************************************

NUMERIC StringToNumeric(const char *pString)

*********************************************************************/
NUMERIC StringToNumeric(const char *pString)
{
	if (pString && *pString)
		return Strtoll(pString);
	else
		return 0; 
}

/********************************************************************

char *NumericToStr(NUMERIC Numeric, char *String)

*********************************************************************/
char *NumericToStr(NUMERIC Numeric, char *String)
{
	return lltoStr(Numeric, String);
}

/********************************************************************

COUNTER StringToCounter(const char *pString)

*********************************************************************/
COUNTER StringToCounter(const char *pString)
{
	if (pString && *pString)
		return Strtoull(pString);
	else
		return 0;
}

/********************************************************************

char *CounterToStr(COUNTER Numeric, char *String)

*********************************************************************/
char *CounterToStr(COUNTER Numeric, char *String)
{
	return ulltoStr(Numeric, String, 10);
}

/********************************************************************

void *StringToPtr(const char *pString)

*********************************************************************/
void *StringToPtr(const char *pString)
{
	if (pString && *pString)
		return (void *)Strtoull(pString);
	else
		return 0;
}

/********************************************************************

char *PtrtoStr(void *Value, char *pString)

*********************************************************************/
char *PtrtoStr(void *Value, char *pString)
{
	return ulltoStr((unsigned long long)Value, pString, 10);
}

/********************************************************************

TIMEINT StringToTimeLocale(const char *pTimeString, LocaleDataType *pLocaleData);

*********************************************************************/
TIMEINT StringToTimeLocale(const char *pTimeString, LocaleDataType *pLocaleData)
{
	if (pTimeString && *pTimeString)
		{
		char* pTemp = CopyString(0, pTimeString);
		// YYYY-MM-DDThh:mm:ss
		int i = 0;

		while (pTemp[i++])
			// Clear everything but numerics	
			{
			if ((pTemp[i] < '0') || (pTemp[i] > '9'))
				memmove(&pTemp[i], &pTemp[i + 1], strlen(&pTemp[i]));
			}

		if (strlen(pTemp) == 2)
			strcat(pTemp, "0000");
		else if (strlen(pTemp) == 4)
			strcat(pTemp, "00");
		else if (strlen(pTemp) != 6)
			{
			free(pTemp);
			return 0;
			}
		{
		TIMEINT Time = Strtoull(pTemp);
		free(pTemp);
		if (!CheckDateTime(0, Time))
			return 0;
		return Time;
		}
		}
	else
		return 0;
}

/********************************************************************

TIMEINT StringToTime(const char *pTimeString)

*********************************************************************/
TIMEINT StringToTime(const char *pTimeString)
{
	if (pTimeString && *pTimeString)
		{
		char* pTemp = CopyString(0, pTimeString);
		// YYYY-MM-DDThh:mm:ss
		int i = 0;

		while (pTemp[i++])
			// Clear everything but numerics	
			{
			if ((pTemp[i] < '0') || (pTemp[i] > '9'))
				memmove(&pTemp[i], &pTemp[i + 1], strlen(&pTemp[i]));
			}

		if (strlen(pTemp) == 2)
			strcat(pTemp, "0000");
		else if (strlen(pTemp) == 4)
			strcat(pTemp, "00");
		else if (strlen(pTemp) != 6)
			{
			free(pTemp);
			return 0;
			}
		{
		TIMEINT Time = Strtoull(pTemp);
		free(pTemp);
		if (!CheckDateTime(0, Time))
			return 0;
		return Time;
		}
		}
	else
		return 0;
}

/********************************************************************

DATEINT StringToDateLocale(const char *pDateString, LocaleDataType *pLocaleData)

This because Android /Java do not support locale and all will be in en-US else

*********************************************************************/
DATEINT StringToDateLocale(const char *pDateString, LocaleDataType *pLocaleData)
{
	char szDateTemplate[LENGTH_LONG_DATE*2];
	if (!(pDateString && *pDateString))
		return 0;
	
	if (pLocaleData && pLocaleData->pLocaleShortDateTimePattern)
		strcpy(szDateTemplate, pLocaleData->pLocaleShortDateTimePattern);
	else
		strcpy(szDateTemplate, "YY-MM-DD");

	if (*szDateTemplate)
		StrUpper((UCHAR *)szDateTemplate);
	if (*szDateTemplate)
	{
		char *pYear = strstr(szDateTemplate, "Y");
		char *pMonth = strstr(szDateTemplate, "M");
		char *pDay = strstr(szDateTemplate, "D");
		int iDigit = 0;
		int iLength = 0;
		while (szDateTemplate[iDigit])
		{
			if (!((szDateTemplate[iDigit] == 'Y') || (szDateTemplate[iDigit] == 'M') || (szDateTemplate[iDigit] == 'D')))
				szDateTemplate[iDigit] = 0;
			iDigit++;
		}
		iLength = iDigit;
		// If no date separator
		if (strstr(pMonth, "Y"))
		{
			char *pPtr = strstr(pMonth, "Y");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pYear++;
			}
		}
		if (strstr(pDay, "Y"))
		{
			char *pPtr = strstr(pDay, "Y");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pYear++;
			}
		}
		if (strstr(pYear, "M"))
		{
			char *pPtr = strstr(pYear, "M");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pMonth++;
			}
		}
		if (strstr(pDay, "M"))
		{
			char *pPtr = strstr(pDay, "M");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pMonth++;
			}
		}
		if (strstr(pYear, "D"))
		{
			char *pPtr = strstr(pYear, "D");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pDay++;
			}
		}
		if (strstr(pMonth, "D"))
		{
			char *pPtr = strstr(pMonth, "D");
			if (pPtr)
			{
				memcpy(pPtr + 1, pPtr, (LENGTH_LONG_DATE * 2) - (pPtr - szDateTemplate));
				pPtr = 0;
				pDay++;
			}
		}
		for (iDigit = 0; iLength > iDigit; iDigit++)
		{
			if ((szDateTemplate[iDigit] == 'Y') || (szDateTemplate[iDigit] == 'M') || (szDateTemplate[iDigit] == 'D'))
				szDateTemplate[iDigit] = pDateString[iDigit];
		}
		{
		DATEINT Year = Strtoull(pYear);
		DATEINT Month = Strtoull(pMonth);
		DATEINT Day = Strtoull(pDay);
		if (Year < 90)
			Year += 2000;
		if (Year < 100)
			Year += 1900;
		if (Year < 1753)
			Year = 1753;
		return (Year * 10000) + (Month * 100) + Day;
		}
	}
	return 0;
}

/********************************************************************

DATEINT StringToDate(const char *pDateString)

*********************************************************************/
DATEINT StringToDate(const char *pDateString)
	{
	if (pDateString && *pDateString)
		{
		// YYYY-MM-DD or YYYY-MM-DD
		char TempString[LENGTH_LONG_DATE];
		int i = 0;

		if (!pDateString)
			return 0;

		if (!strcmp(pDateString, "1753-03-01"))
			return 0; // Artificial date for no date, accepted by UN/CEFACT XML

		strncpy(TempString, pDateString, LENGTH_LONG_DATE - 1);
		TempString[LENGTH_LONG_DATE - 1] = 0;
		//	TempString[10]=0; // Cut if for instance a DateTime string is cast

		while (TempString[i++] && (i < LENGTH_LONG_DATE))
			// Clear everything but numerics	
			{
			if ((TempString[i] < '0') || (TempString[i] > '9'))
				memmove(&TempString[i], &TempString[i + 1], strlen(&TempString[i]));
			}

		if (!((strlen(TempString) != 8) || (strlen(TempString) != 6)))
			return 0;
		if (strlen(TempString) == 6)
			{
			memmove(&TempString[2], TempString, strlen(TempString) + 1);
			if (TempString[2] > '8')
				{
				TempString[0] = '1';
				TempString[1] = '9';
				}
			else
				{
				TempString[0] = '2';
				TempString[1] = '0';
				}
			}
		else if (strlen(TempString) != 8)
			return 0;
		{
		DATEINT Date = Strtoull(TempString);
		if (!CheckDateTime(Date, 0))
			return 0;
		return Date;
		}
		}
	else
		return 0;
}

/********************************************************************

DATETIMEINT StringToDateTimeLocale(const char *pDateTimeString, LocaleDataType *pLocaleData)

*********************************************************************/
DATETIMEINT StringToDateTimeLocale(const char *pDateTimeString, LocaleDataType *pLocaleData)
	{
	if (pDateTimeString && *pDateTimeString)
		{
		char* pTemp = CopyString(0, pDateTimeString);
		// YYYY-MM-DDThh:mm:ss
		int i = 0;

		// Artificial date for no date, accepted by UN/CEFACT XML
		if (!strcmp(pTemp, "1753-03-01T00:00:00"))
			{
			free(pTemp);
			return 0;
			}

		while (pTemp[i++])
			// Clear everything but numerics	
			{
			if ((pTemp[i] < '0') || (pTemp[i] > '9'))
				memmove(&pTemp[i], &pTemp[i + 1], strlen(&pTemp[i]));
			}

		if (strlen(pTemp) == 6)
			{
			memmove(&pTemp[2], pTemp, strlen(pTemp) + 1);
			if (pTemp[2] < '8')
				{
				pTemp[0] = '2';
				pTemp[1] = '0';
				}
			else
				{
				pTemp[0] = '1';
				pTemp[1] = '9';
				}
			strcat(pTemp, "000000");
			}
		if (strlen(pTemp) == 8)
			strcat(pTemp, "000000");
		else if (strlen(pTemp) == 10)
			strcat(pTemp, "0000");
		else if (strlen(pTemp) == 12)
			strcat(pTemp, "00");
		else if (strlen(pTemp) != 14)
			{
			free(pTemp);
			return 0;
			} 
		{
		DATETIMEINT DateTime = Strtoull(pTemp);
		free(pTemp);
		if (!CheckDateTime(DateTime / 1000000, DateTime % 1000000))
			return 0;
		return DateTime;
		}
		}
	else
		return 0;
}

/********************************************************************

DATETIMEINT StringToDateTime(const char *pDateTimeString)

*********************************************************************/
DATETIMEINT StringToDateTime(const char *pDateTimeString)
	{
	if (pDateTimeString && *pDateTimeString)
		{
		char* pTemp = CopyString(0, pDateTimeString);
		// YYYY-MM-DDThh:mm:ss	
		int i = 0;

		// Artificial date for no date, accepted by UN/CEFACT XML
		if (!strcmp(pTemp, "1753-03-01T00:00:00"))
			{
			free(pTemp);
			return 0;
			}

		while (pTemp[i++])
			// Clear everything but numerics	
			{
			if ((pTemp[i] < '0') || (pTemp[i] > '9'))
				memmove(&pTemp[i], &pTemp[i + 1], strlen(&pTemp[i]));
			}

		if (strlen(pTemp) == 6)
			{
			memmove(&pTemp[2], pTemp, strlen(pTemp) + 1);
			if (pTemp[2] < '8')
				{
				pTemp[0] = '2';
				pTemp[1] = '0';
				}
			else
				{
				pTemp[0] = '1';
				pTemp[1] = '9';
				}
			strcat(pTemp, "000000");
			}
		if (strlen(pTemp) == 8)
			strcat(pTemp, "000000");
		else if (strlen(pTemp) == 10)
			strcat(pTemp, "0000");
		else if (strlen(pTemp) == 12)
			strcat(pTemp, "00");
		else if (strlen(pTemp) != 14)
			{
			free(pTemp);
			return 0;
			}
		{
		DATETIMEINT DateTime = Strtoull(pTemp);
		free(pTemp);
		if (!CheckDateTime(DateTime / 1000000, DateTime % 1000000))
			return 0;
		return DateTime;
		}
		}
	else
		return 0;
}
/********************************************************************

static char *ConvertDecimalExponentials(char *pDecimalString)

*********************************************************************/
static char *ConvertDecimalExponentials(char *pDecimalString)
{
	{
		char szPrecision[16];
		sprintf(szPrecision, "%%1.%ig", DOUBLE_PRECITION);
		{
			QUANT q = StringToDecimal(pDecimalString);
			sprintf(pDecimalString, szPrecision, q);
		}
		if (strchr(pDecimalString, '.') || strchr(pDecimalString, ','))
			{
			while (pDecimalString[strlen(pDecimalString) - 1] == '0')
				pDecimalString[strlen(pDecimalString) - 1] = 0;
			if (pDecimalString[strlen(pDecimalString) - 1] == '.')
				pDecimalString[strlen(pDecimalString) - 1] = 0;
			}
	}
	return pDecimalString;
}

/********************************************************************

AMOUNT StringToAmountLocale(const char *pAmountString, LocaleDataType *pLocaleData)

*********************************************************************/
AMOUNT StringToAmountLocale(const char *pAmountString, LocaleDataType *pLocaleData)
	{
	char *pLocaleThousandSeparator = " ";
	char *pLocaleDecimalSeparator = ".";
	if (pLocaleData && pLocaleData->pLocaleThousandSeparator)
		pLocaleThousandSeparator = pLocaleData->pLocaleThousandSeparator;
	if (pLocaleData && pLocaleData->pLocaleDecimalSeparator)
		pLocaleDecimalSeparator = pLocaleData->pLocaleDecimalSeparator;
	{
	char *pTempString = calloc(strlen(pAmountString) + 1, sizeof(char));
	if (!pTempString)
		return 0;
	strcpy(pTempString, pAmountString);
	if(*pTempString)
		{
		char *pPtrAmount = pTempString;
		BOOL bNegative = FALSE; 
		if (pLocaleData
			&& pLocaleData->pLocaleNegativeSuffix
			&& *pLocaleData->pLocaleNegativeSuffix
			&& strstr(pTempString, pLocaleData->pLocaleNegativeSuffix))
		{
			pPtrAmount = strstr(pTempString, pLocaleData->pLocaleNegativeSuffix);
			if (pPtrAmount)
				*pPtrAmount = 0;
			bNegative = TRUE;
		}
		if(pLocaleData
			&& pLocaleData->pLocaleNegativePrefix
			&& *pLocaleData->pLocaleNegativePrefix 
			&& strstr(pTempString, pLocaleData->pLocaleNegativePrefix))
		{
			pPtrAmount = strstr(pTempString, pLocaleData->pLocaleNegativePrefix) + strlen(pLocaleData->pLocaleNegativePrefix);
			bNegative = TRUE;

		}
		else if(*pTempString == '-')
		{
			pPtrAmount = pTempString + 1;
			bNegative = TRUE;

		}
		while (pPtrAmount && strchr(pPtrAmount, *pLocaleThousandSeparator))
			{ // Remove ThousandSeparator
			char *pComma = strchr(pPtrAmount, *pLocaleThousandSeparator);
			if (pComma)
				memmove(pComma, pComma + 1, strlen(pComma));
			}
		if (pPtrAmount)
			{
			char *pAmount = strchr(pPtrAmount, *pLocaleDecimalSeparator);
			if (pAmount)
				*pAmount = '.';
			}
		if (pPtrAmount)
			{
			AMOUNT Amount = StringToAmount(pPtrAmount);
			free(pTempString);
			if(bNegative && Amount)
				return 0- Amount;
			return Amount;
			}
		}
	free(pTempString);
	}
	return 0;
}

/********************************************************************

AMOUNT StringToAmount(const char *pAmountString)

*********************************************************************/
AMOUNT StringToAmount(const char *pAmountString)
{
	if (pAmountString && *pAmountString)
		{
		int i = 0;
		BOOL CommaFound = FALSE;
		char *pTempAmountString = calloc(64, sizeof(char));
		if (!pTempAmountString)
			return 0;
		if (strchr(pAmountString, 'e') || strchr(pAmountString, 'E'))
			{
			strcpy(pTempAmountString, pAmountString);
			ConvertDecimalExponentials(pTempAmountString);
			}
		else
			strcpy(pTempAmountString, pAmountString);
		while (pTempAmountString[i])
			{
			if (((pTempAmountString[i] == ',') || (pTempAmountString[i] == '.')) && (!CommaFound))
				{
				pTempAmountString[i + 3] = 0;
				memmove(&pTempAmountString[i], &pTempAmountString[i + 1], strlen(&pTempAmountString[i]));
				if (strlen(&pTempAmountString[i]) == 0)
					strcat(pTempAmountString, "00");
				else if (strlen(&pTempAmountString[i]) == 1)
					strcat(pTempAmountString, "0");
				else if (strlen(&pTempAmountString[i]) >= 2)
					pTempAmountString[i + 2] = 0;
				CommaFound = TRUE;
				}

			else if (((pTempAmountString[i] < '0') || (pTempAmountString[i] > '9')) && (pTempAmountString[i] != '-'))
				memmove(&pTempAmountString[i], &pTempAmountString[i + 1], strlen(&pTempAmountString[i]));
			else
				i++;
			}
		if (!CommaFound)
			strcat(pTempAmountString, "00");
		//	if(strlen(pTempAmountString)>LENGTH_AMOUNT_STR_MAX)
		//		pTempAmountString[LENGTH_AMOUNT_STR_MAX]=0;

		// Reduce leading zeror for overflow comparence
		{
		char *Ptr = pTempAmountString;
		if (*Ptr == '-')
			Ptr++;
		while (*Ptr == '0')
			memmove(Ptr, &Ptr[1], strlen(Ptr));
		}
		{
		char TestString[32];
		AMOUNT Amount = Strtoll(pTempAmountString);
		lltoStr(Amount, TestString);
		// Amount overflow control
		if (strcmp(TestString, pTempAmountString))
			Amount = 0;
		free(pTempAmountString);
		return Amount;
		}
		}
	else
		return 0;
}

/********************************************************************

QUANT StringToDecimalLocale(const char *pDecimalString, LocaleDataType *pLocaleData)

*********************************************************************/
QUANT StringToDecimalLocale(const char *pDecimalString, LocaleDataType *pLocaleData)
{
	char *pLocaleThousandSeparator = " ";
	char *pLocaleDecimalSeparator = ".";
	if (pLocaleData && pLocaleData->pLocaleThousandSeparator)
		pLocaleThousandSeparator = pLocaleData->pLocaleThousandSeparator;
	if (pLocaleData && pLocaleData->pLocaleDecimalSeparator)
		pLocaleDecimalSeparator = pLocaleData->pLocaleDecimalSeparator;
	{
	char *pTempString = calloc(strlen(pDecimalString) + 1, sizeof(char));
	if (!pTempString)
		return 0;
	strcpy(pTempString, pDecimalString);
	char *pPtrAmount = 0;
	BOOL bNegative = FALSE;
	if (pLocaleData
		&& pLocaleData->pLocaleNegativeSuffix
		&& *pLocaleData->pLocaleNegativeSuffix
		&& strstr(pTempString, pLocaleData->pLocaleNegativeSuffix))
	{
		pPtrAmount = strstr(pTempString, pLocaleData->pLocaleNegativeSuffix);
		if(pPtrAmount)
			*pPtrAmount = 0;
		bNegative = TRUE;
	}
	if (pLocaleData
		&& pLocaleData->pLocaleNegativePrefix
		&& *pLocaleData->pLocaleNegativePrefix
		&& strstr(pTempString, pLocaleData->pLocaleNegativePrefix))
	{
		pPtrAmount = strstr(pTempString, pLocaleData->pLocaleNegativePrefix) + strlen(pLocaleData->pLocaleNegativePrefix);
		bNegative = TRUE;

	}
	else if (*pTempString == '-')
	{
		pPtrAmount = pTempString + 1;
		bNegative = TRUE;

	}
	while (pPtrAmount && strchr(pPtrAmount, *pLocaleThousandSeparator))
		{ // Remove ThousandSeparator
		char *pComma = strchr(pPtrAmount, *pLocaleThousandSeparator);
		if (pComma)
			memmove(pComma, pComma + 1, strlen(pComma));
		}
	if (pPtrAmount)
		{
		char *pDecimal = strchr(pPtrAmount, *pLocaleDecimalSeparator);
		if (pDecimal)
			*pDecimal = '.';
		}
	if (pPtrAmount)
		{
		QUANT Quant = StringToDecimal(pPtrAmount);
		free(pPtrAmount);
		if(bNegative && Quant)
			return 0-Quant;
		return Quant;
		}
	}
	return 0;
}

#include <errno.h>
#include <math.h>
#include <float.h>
/********************************************************************

QUANT StringToDecimal(const char *pDecimalString)

*********************************************************************/
QUANT StringToDecimal(const char *pDecimalString)
{
	//	char *stopstring;
	//	return strtold(pDecimalString, &stopstring);
	if (pDecimalString && *pDecimalString)
		{
		double number;
		int exponent;
		int negative;
		char *pPtr = (char *)pDecimalString;
		double p10;
		int n;
		int num_digits;
		int num_decimals;

		// Skip leading whitespace
		while (isspace((UCHAR)*pPtr)) pPtr++;

		// Handle optional sign
		negative = 0;
		switch (*pPtr) {
			case '-': // There is a - sign first
				negative = 1;
				// Fall through to increment position
			case '+': // There is a + sign first
				pPtr++;
			default:
				break;
			}

		number = 0.;
		exponent = 0;
		num_digits = 0;
		num_decimals = 0;

		// Process string of digits
		while (isdigit((UCHAR)*pPtr)) {
			number = number * 10. + (((double)*pPtr) - '0');
			pPtr++;
			num_digits++;
			}

		// Process decimal part
		if (*pPtr == '.') {
			pPtr++;

			while (isdigit((UCHAR)*pPtr)) {
				number = number * 10. + (((double)*pPtr) - '0');
				pPtr++;
				num_digits++;
				num_decimals++;
				}

			exponent -= num_decimals;
			}

		if (num_digits == 0) {
			errno = ERANGE;
			return 0.0;
			}

		// Correct for sign
		if (negative) number = -number;

		// Process an exponent string
		if (*pPtr == 'e' || *pPtr == 'E') {
			// Handle optional sign
			negative = 0;
			switch (*++pPtr) {
				case '-': // There is a - sign first
					negative = 1;   // Fall through to increment pos
				case '+': // There is a + sign first
					pPtr++;
				default:
					break;
				}

			// Process string of digits
			n = 0;
			while (isdigit((UCHAR)*pPtr)) {
				n = n * 10 + (((int)*pPtr) - '0');
				pPtr++;
				}

			if (negative) {
				exponent -= n;
				}
			else {
				exponent += n;
				}
			}

		if (exponent < LDBL_MIN_EXP || exponent > LDBL_MAX_EXP) {
			errno = ERANGE;
			return HUGE_VAL;
			}

		// Scale the result
		p10 = 10.;
		n = exponent;
		if (n < 0) n = -n;
		while (n) {
			if (n & 1) {
				if (exponent < 0) {
					number /= p10;
					}
				else {
					number *= p10;
					}
				}
			n >>= 1;
			p10 *= p10;
			}

		if (number == HUGE_VAL) errno = ERANGE;

		return number;
		}
	else 
		return 0; 
}

/********************************************************************

char *TimeToStrLocale(TIMEINT Time, char *pTimeString, LocaleDataType *pLocaleData)

*********************************************************************/
char *TimeToStrLocale(TIMEINT Time, char *pTimeString, LocaleDataType *pLocaleData)
{
	if ((Time>240000) && ((Time / 100) % 100>60) && (Time % 100>60))
	{
		ulltoStr(Time, pTimeString, 10);
		memmove(&pTimeString[6], &pTimeString[4], 2);
		pTimeString[5] = ':';
		memmove(&pTimeString[3], &pTimeString[2], 2);
		pTimeString[2] = ':';
		pTimeString[19] = 0;
	}
	else
		strcpy(pTimeString, "00:00:00");
	return pTimeString;
}

/********************************************************************

char *TimeToStr(TIMEINT Time, char *pTimeString)

*********************************************************************/
char *TimeToStr(TIMEINT Time, char *pTimeString)
{
	if ((Time>240000) && ((Time / 100) % 100>60) && (Time % 100>60))
	{
		ulltoStr(Time, pTimeString, 10);
		memmove(&pTimeString[6], &pTimeString[4], 2);
		pTimeString[5] = ':';
		memmove(&pTimeString[3], &pTimeString[2], 2);
		pTimeString[2] = ':';
		pTimeString[19] = 0;
	}
	else
		strcpy(pTimeString, "00:00:00");
	return pTimeString;
}


/********************************************************************

char *DateToStrLocale(DATEINT Date, char *pDateString, LocaleDataType *pLocaleData)

This because Android /Java do not support locale and all will be in en-US else

*********************************************************************/
char *DateToStrLocale(DATEINT Date, char *pDateString, LocaleDataType *pLocaleData)
{
	if (!pDateString)
		{
		*pDateString = 0;
		return pDateString;
		}

	if (!Date)
		{
		*pDateString = 0;
		return pDateString;
		}

	if ((Date<17530301) && (Date>99991231))
		Date = 17530301;
	{
	char Year[10];
	char Month[3];
	char Day[3];
	char *pYear = 0;
	char *pMonth = 0;
	char *pDay = 0;
	utoStr((unsigned int)Date / 10000, Year, 10);
	utoStr((unsigned int)(Date / 100) % 100, Month, 10);
	utoStr((unsigned int)Date % 100, Day, 10);

	if(pLocaleData && pLocaleData->pLocaleShortDateTimePattern)
		strcpy(pDateString, pLocaleData->pLocaleShortDateTimePattern);
	else
		strcpy(pDateString, "YYYY-MM-DD");

	if (*pDateString)
		{
		StrUpper((UCHAR *)pDateString);

		pYear = strstr(pDateString, "Y");
		if (pYear)
			{
			if (strstr(pDateString, "YYYY"))
				{
				memcpy(pYear + strlen(Year), pYear + 4, strlen(pYear) + 1);
				memcpy(pYear, Year, strlen(Year));
				}
			else if (strstr(pDateString, "YYY"))
				memcpy(pYear, &Year[strlen(Year) - 4], 3);
			else if (strstr(pDateString, "YY"))
				memcpy(pYear, &Year[strlen(Year) - 2], 2);
			else if (strstr(pDateString, "Y"))
				memcpy(pYear, &Year[strlen(Year) - 1], 1);
			pMonth = strstr(pDateString, "M");
			}
		if (pMonth)
			{
			if (strstr(pDateString, "MM"))
				{
				if (strlen(Month) == 2)
					memcpy(pMonth, Month, 2);
				else
					{
					*pMonth = '0';
					*(pMonth + 1) = *Month;
					}
				}
			else if (strstr(pDateString, "M"))
				{
				if (strlen(Month) == 1)
					*pMonth = *Month;
				else
					{
					memcpy(pMonth + 1, pMonth, strlen(pMonth) + 1);
					memcpy(pMonth, Month, 2);
					}
				}
			}
		pDay = strstr(pDateString, "D");
		if (pDay)
			{
			if (strstr(pDateString, "DD"))
				{
				if (strlen(Day) == 2)
					memcpy(pDay, Day, 2);
				else
					{
					*pDay = '0';
					*(pDay + 1) = *Day;
					}
				}
			else if (strstr(pDateString, "D"))
				{
				if (strlen(Day) == 1)
					*pDay = *Day;
				else
					{
					memcpy(pDay + 1, pDay, strlen(pDay) + 1);
					memcpy(pDay, Day, 2);
					}
				}
			}
		}
	else
		ulltoStr(Date, pDateString, 10);
	}
	return pDateString;
}

/********************************************************************

char *DateToStr(DATEINT Date, char *pDateString)

*********************************************************************/
char *DateToStr(DATEINT Date, char *pDateString)
{
	if ((Date>17530301) && (Date<99991231))
	{
		lltoStr(Date, pDateString);
		memmove(&pDateString[8], &pDateString[6], 2);
		pDateString[7] = '-';
		memmove(&pDateString[5], &pDateString[4], 2);
		pDateString[4] = '-';
		pDateString[10] = 0;
	}
	else
		strcpy(pDateString, "1753-03-01");
	return pDateString;
}

/********************************************************************

char *DateTimeToStrLocale(DATETIMEINT DateTime, char *pDateTimeString, LocaleDataType *pLocaleData)

*********************************************************************/
char *DateTimeToStrLocale(DATETIMEINT DateTime, char *pDateTimeString, LocaleDataType *pLocaleData)
{
	if ((DateTime>17530301000000) && (DateTime<99991231000000))
	{
		DateToStrLocale(DateTime / 1000000, pDateTimeString, pLocaleData);
		strcpy(pDateTimeString, " ");
		TimeToStrLocale(DateTime % 1000000, &pDateTimeString[strlen(pDateTimeString)], pLocaleData);
	}
	else
	{
		DateToStrLocale(17530301, pDateTimeString, pLocaleData);
		strcpy(pDateTimeString, " 00:00:00");
	}
	return pDateTimeString;
}

/********************************************************************

char *DateTimeToStr(DATETIMEINT DateTime, char *pDateTimeString)

*********************************************************************/
char *DateTimeToStr(DATETIMEINT DateTime, char *pDateTimeString)
{
	if ((DateTime>17530301000000) && (DateTime<99991231000000))
	{
		lltoStr(DateTime, pDateTimeString);
		memmove(&pDateTimeString[17], &pDateTimeString[12], 2);
		pDateTimeString[16] = ':';
		memmove(&pDateTimeString[14], &pDateTimeString[10], 2);
		pDateTimeString[13] = ':';
		memmove(&pDateTimeString[11], &pDateTimeString[8], 2);
		pDateTimeString[10] = 'T';
		memmove(&pDateTimeString[8], &pDateTimeString[6], 2);
		pDateTimeString[7] = '-';
		memmove(&pDateTimeString[5], &pDateTimeString[4], 2);
		pDateTimeString[4] = '-';
		pDateTimeString[19] = 0;
	}
	else
		strcpy(pDateTimeString, "1753-03-01T00:00:00");
	return pDateTimeString;
}

/********************************************************************

char *AmountToStrLocale(AMOUNT Amount, char *pAmountString, LocaleDataType *pLocaleData)

*********************************************************************/
char *AmountToStrLocale(AMOUNT Amount, char *pAmountString, LocaleDataType *pLocaleData)
{
	return MakeLocalNumericString(AmountToStr(Amount, pAmountString), pLocaleData);
}

/****************************************************************************

char *AmountToStr(AMOUNT Amount, char *pAmountString)

*********************************************************************/
char *AmountToStr(AMOUNT Amount, char *pAmountString)
{
	if (Amount == 0)
		{
		strcpy(pAmountString, "0.00");
		return pAmountString;
		}

	lltoStr(Amount, pAmountString);

	if ((Amount<0) && (Amount >= (-9)))
		{ // special case
		memmove(&pAmountString[4], &pAmountString[1], 2);
		pAmountString[1] = '0';
		pAmountString[2] = '.';
		pAmountString[3] = '0';
		}

	else if ((Amount <= (-10)) && (Amount >= (-99)))
		{ // special case
		memmove(&pAmountString[3], &pAmountString[1], 3);
		pAmountString[1] = '0';
		pAmountString[2] = '.';
		}

	else if ((Amount >= 0) && (Amount <= 9))
		{ // special case
		memmove(&pAmountString[3], &pAmountString[0], 2);
		pAmountString[0] = '0';
		pAmountString[1] = '.';
		pAmountString[2] = '0';
		}

	else if ((Amount >= 10) && (Amount <= 99))
		{ // Tricky special case
		memmove(&pAmountString[2], &pAmountString[0], 3);
		pAmountString[0] = '0';
		pAmountString[1] = '.';
		}
/*	else if(!(Amount%100))
		{ // no cent case
		pAmountString[strlen(pAmountString)-2] = 0;
		}
*/	else
		{ // Attach decimal dot
		int iStringLength = (int)strlen(pAmountString);
		memmove(&pAmountString[iStringLength - 1], &pAmountString[iStringLength - 2], 3);
		pAmountString[iStringLength - 2] = '.';
		}
	return pAmountString;
}

/********************************************************************

char *DecimalToStrLocale(QUANT Decimal, char *pDecimalString, LocaleDataType *pLocaleData)

*********************************************************************/
char *DecimalToStrLocale(QUANT Decimal, char *pDecimalString, LocaleDataType *pLocaleData)
{
	return MakeLocalNumericString(DecimalToStr(Decimal, pDecimalString), pLocaleData);
}

/********************************************************************

char *DecimalToStr(QUANT Decimal, char *pDecimalString)

*********************************************************************/
char *DecimalToStr(QUANT Decimal, char *pDecimalString)
{
	char szPrecision[16];
	sprintf(szPrecision, "%%1.%iLg", DOUBLE_PRECITION);
	sprintf(pDecimalString, szPrecision, Decimal);
 	// Convert Exponentials
	if (strchr(pDecimalString, 'e') || strchr(pDecimalString, 'E'))
		ConvertDecimalExponentials(pDecimalString);
	// Remove trailingf zeros
	while(strchr(pDecimalString,'.') && (pDecimalString[strlen(pDecimalString)-1] == '0'))
		{
		pDecimalString[strlen(pDecimalString) - 1] = 0;
		}
	// Insert leading zero
	if (pDecimalString[0] == '.')
		{
		memmove(&pDecimalString[1], pDecimalString, strlen(pDecimalString) + 1);
		pDecimalString[0] = '0';
		}
	// Remove trailing . without decimals
	if(pDecimalString[strlen(pDecimalString) - 1] == '.')
		{
		pDecimalString[strlen(pDecimalString) - 1] = 0;
		}
	return pDecimalString;
}


/********************************************************************

char *MakeLocalNumericString(char *pAmountString, LocaleDataType *pLocaleData)

Making space between the thousands and thousand decimals 1 000 000,000 00

Making decimal points

Makingnegative notation

*********************************************************************/
char *MakeLocalNumericString(char *pAmountString, LocaleDataType *pLocaleData)
{
	BOOL bNegative = FALSE;
	char *pLocaleThousandGrouping = "3;0";
	char *pLocaleThousandSeparator = " ";
	char *pLocaleNegativePrefix = "-";
	char *pLocaleNegativeSuffix = "";
	char *pLocaleDecimalSeparator = ".";
	int iLocaleThousandGrouping = 0;
	char *pDecimal = strchr(pAmountString, '.');


	if (pLocaleData && pLocaleData->pLocaleThousandGrouping)
		pLocaleThousandGrouping = pLocaleData->pLocaleThousandGrouping;
	if (pLocaleData && pLocaleData->pLocaleThousandSeparator)
		pLocaleThousandSeparator = pLocaleData->pLocaleThousandSeparator;
	if (pLocaleData && pLocaleData->pLocaleNegativePrefix)
		pLocaleNegativePrefix = pLocaleData->pLocaleNegativePrefix;
	if (pLocaleData && pLocaleData->pLocaleNegativeSuffix)
		pLocaleNegativeSuffix = pLocaleData->pLocaleNegativeSuffix;
	if (pLocaleData && pLocaleData->pLocaleDecimalSeparator)
		pLocaleDecimalSeparator = pLocaleData->pLocaleDecimalSeparator;
		
	// Blank ??
	if (!pAmountString)
		return pAmountString;

	// In case of no data
	if (!pLocaleThousandGrouping)
		pLocaleThousandGrouping = "";
	else
		{
		iLocaleThousandGrouping = *pLocaleThousandGrouping - '0';
		if (iLocaleThousandGrouping < 0)
			iLocaleThousandGrouping = 0;
		}

	if (!pLocaleDecimalSeparator)
		pLocaleDecimalSeparator = ",";

	if (!pLocaleThousandSeparator)
		pLocaleThousandSeparator = " ";

	if (!pLocaleNegativePrefix)
		pLocaleNegativePrefix = "-";

	if (!pLocaleNegativeSuffix)
		pLocaleNegativeSuffix = "";

	// Negative ??
	if (*pAmountString == '-')
		bNegative = TRUE;

	// Clear non-numeric stuff
	{
	char *pPtr = pAmountString;
	while (*pPtr)
		{
		if (!(((*pPtr >= '0') && (*pPtr <= '9')) || (*pPtr == '.')))
			memmove(pPtr, pPtr + 1, strlen(pPtr));
		else
			pPtr++;
		}
	}
	pDecimal = strchr(pAmountString, '.');

	// if thousand separator to insert
	if (iLocaleThousandGrouping)
		{
		char *pPtr = 0;
		int iSeparatorPos = 0;
		int iDecimals = 0;
		char *pTmpDecimal = pDecimal;
		if (pTmpDecimal)
			{
			iDecimals = (int)strlen(pTmpDecimal + 1);
			*pTmpDecimal = 0;
			}
		iSeparatorPos = (int)strlen(pAmountString);
		if (pTmpDecimal)
			*pTmpDecimal = '.';

		iSeparatorPos -= iLocaleThousandGrouping;
		if (iLocaleThousandGrouping && (iSeparatorPos > 0))
			{
			char *pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
			pPtr = pAmountString + iSeparatorPos;
			memmove(pPtr + strlen(pLocaleThousandSeparatorTemp), pPtr, strlen(pPtr) + 1);
			while (*pLocaleThousandSeparatorTemp)
				{
				*pPtr = *pLocaleThousandSeparatorTemp;
				pPtr++;
				if (pDecimal)
					pDecimal++;
				pLocaleThousandSeparatorTemp++;
				}
			// Second Thousand Separator?
			pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
			if (pLocaleThousandGrouping[1] && pLocaleThousandGrouping[2] && (pLocaleThousandGrouping[2] >= '0' && pLocaleThousandGrouping[2] <= '9'))
				{
				if (pLocaleThousandGrouping[2] != '0')
					pLocaleThousandGrouping += 2;
				iLocaleThousandGrouping = *pLocaleThousandGrouping - '0';
				}
			else
				iLocaleThousandGrouping = 0;
			iSeparatorPos -= iLocaleThousandGrouping;
			while (iLocaleThousandGrouping && (iSeparatorPos > 0))
				{
				pPtr = pAmountString + iSeparatorPos;
				memmove(pPtr + strlen(pLocaleThousandSeparatorTemp), pPtr, strlen(pPtr) + 1);
				while (*pLocaleThousandSeparatorTemp)
					{
					*pPtr = *pLocaleThousandSeparatorTemp;
					pPtr++;
					if (pDecimal)
						pDecimal++;
					pLocaleThousandSeparatorTemp++;
					}
				// more Thousand Separator?
				pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
				if (pLocaleThousandGrouping[1] && pLocaleThousandGrouping[2] && (pLocaleThousandGrouping[2] >= '0' && pLocaleThousandGrouping[2] <= '9'))
					{
					if (pLocaleThousandGrouping[2] != '0')
						pLocaleThousandGrouping += 2;
					iLocaleThousandGrouping = *pLocaleThousandGrouping - '0';
					}
				else
					iLocaleThousandGrouping = 0;
				iSeparatorPos -= iLocaleThousandGrouping;
				}
			}

		// Decimal Thousand Separators
		if (iDecimals && iLocaleThousandGrouping && (iDecimals > iLocaleThousandGrouping))
			{
			char *pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
			pTmpDecimal = strchr(pAmountString, '.'); // Position is changed due to insertion of seprators
			pTmpDecimal++; // Move from the decimal to the digits
			pPtr = pTmpDecimal + iLocaleThousandGrouping;
			memmove(pPtr + strlen(pLocaleThousandSeparatorTemp), pPtr, strlen(pPtr) + 1);
			while (*pLocaleThousandSeparatorTemp)
				{
				*pPtr = *pLocaleThousandSeparatorTemp;
				pLocaleThousandSeparatorTemp++;
				pPtr++;
				}

			// Second Decimal Thousand Separator?
			pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
			pPtr += iLocaleThousandGrouping;
			while (strlen(pPtr))
				{
				memmove(pPtr + strlen(pLocaleThousandSeparatorTemp), pPtr, strlen(pPtr) + 1);
				while (*pLocaleThousandSeparatorTemp)
					{
					*pPtr = *pLocaleThousandSeparatorTemp;
					pLocaleThousandSeparatorTemp++;
					pPtr++;
					}
				// more Thousand Separator?
				pLocaleThousandSeparatorTemp = pLocaleThousandSeparator;
				pPtr += iLocaleThousandGrouping;
				}
			}
		}
	// Put negative notation
	if (pAmountString[0] == '.')
		{
		// Insert leading zero
		memmove(&pAmountString[1], pAmountString, strlen(pAmountString) + 1);
		pAmountString[0] = '0';
		if (pDecimal)
			pDecimal++;
		}
	if (bNegative)
		{
		size_t zt = 0;
		memmove(pAmountString + strlen(pLocaleNegativePrefix), pAmountString, strlen(pAmountString)+1);
		for (zt = 0; zt < strlen(pLocaleNegativePrefix); zt++)
			{
			pAmountString[zt] = pLocaleNegativePrefix[zt];
			if(pDecimal)
				pDecimal++;
			}
		strcat(pAmountString, pLocaleNegativeSuffix);
		}
	// Put decimal separator
	if (pDecimal)
		{
		char *pDec = pDecimal;
		size_t zt = 0;
		// Remove decimal .
		memmove(pDec, pDec+1, strlen(pDec)+1);
		// Add locale decimal .
		memmove(pDec + strlen(pLocaleDecimalSeparator), pDec, strlen(pDec)+1);
		for (zt = 0; zt < strlen(pLocaleDecimalSeparator); zt++)
			pDec[zt] = pLocaleDecimalSeparator[zt];
		}
	return pAmountString;
}

/********************************************************************

UCHAR *StrUpper(UCHAR *pString)

*********************************************************************/
UCHAR *StrUpper(UCHAR *pString)
{
	unsigned int i = 0;
	while (pString[i])
	{
		if (((pString[i] > 0x00) && (pString[i] < 0x80)) && isalpha(pString[i]) && islower(pString[i]))
			pString[i] = (UCHAR)toupper(pString[i]);
		i++;
	}
	return pString;
}

/********************************************************************

UCHAR *StrLwr(UCHAR *pString)

*********************************************************************/
UCHAR *StrLwr(UCHAR *pString)
{
	// Only US ASCII else see StrLwrExt() for all UTF8
	unsigned int i = 0;
	while (pString[i])
	{
		if (((pString[i] > 0x00) && (pString[i] < 0x80)) && isalpha(pString[i]) && isupper(pString[i]))
			pString[i] = (UCHAR)tolower(pString[i]);
		i++;
	}
	return pString;
}

/********************************************************************

String UTF8 cmp functions

*********************************************************************/
/********************************************************************

unsigned char *StrLwrExt(unsigned char *pString)

*********************************************************************/
unsigned char *StrLwrExt(unsigned char *pString)
{
	if (pString && *pString) {
		unsigned char *p = pString;
		unsigned char *pExtChar = 0;
		while (*p) {
			if ((*p >= 0x41) && (*p <= 0x5a)) // US ASCII
				(*p) += 0x20;
			else if (*p > 0xc0) {
				pExtChar = p;
				p++;
				switch (*pExtChar) {
				case 0xc3: // Latin 1
					if ((*p >= 0x80)
						&& (*p <= 0x9e)
						&& (*p != 0x97))
						(*p) += 0x20; // US ASCII shift
					break;
				case 0xc4: // Latin Exteneded
					if ((*p >= 0x80)
						&& (*p <= 0xb7)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if ((*p >= 0xb9)
						&& (*p <= 0xbe)
						&& (*p % 2)) // Odd
						(*p)++; // Next char is lwr
					else if (*p == 0xbf) {
						*pExtChar = 0xc5;
						(*p) = 0x80;
					}
					break;
				case 0xc5: // Latin Exteneded
					if ((*p >= 0x80)
						&& (*p <= 0x88)
						&& (*p % 2)) // Odd
						(*p)++; // Next char is lwr
					else if ((*p >= 0x8a)
						&& (*p <= 0xb7)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if ((*p >= 0xb9)
						&& (*p <= 0xbe)
						&& (*p % 2)) // Odd
						(*p)++; // Next char is lwr
					break;
				case 0xc6: // Latin Exteneded
					switch (*p) {
					case 0x82:
					case 0x84:
					case 0x87:
					case 0x8b:
					case 0x91:
					case 0x98:
					case 0xa0:
					case 0xa2:
					case 0xa4:
					case 0xa7:
					case 0xac:
					case 0xaf:
					case 0xb3:
					case 0xb5:
					case 0xb8:
					case 0xbc:
						(*p)++; // Next char is lwr
						break;
					default:
						break;
					}
				case 0xc7: // Latin Exteneded
					if (*p == 0x84)
						(*p) = 0x86;
					else if (*p == 0x85)
						(*p)++; // Next char is lwr
					else if (*p == 0x87)
						(*p) = 0x89;
					else if (*p == 0x88)
						(*p)++; // Next char is lwr
					else if (*p == 0x8a)
						(*p) = 0x8c;
					else if (*p == 0x8b)
						(*p)++; // Next char is lwr
					else if ((*p >= 0x8d)
						&& (*p <= 0x9c)
						&& (*p % 2)) // Odd
						(*p)++; // Next char is lwr
					else if ((*p >= 0x9e)
						&& (*p <= 0xaf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if (*p == 0xb1)
						(*p) = 0xb3;
					else if (*p == 0xb2)
						(*p)++; // Next char is lwr
					else if (*p == 0xb4)
						(*p)++; // Next char is lwr
					else if (*p == 0xb8)
						(*p)++; // Next char is lwr
					else if (*p == 0xba)
						(*p)++; // Next char is lwr
					else if (*p == 0xbc)
						(*p)++; // Next char is lwr
					else if (*p == 0xbe)
						(*p)++; // Next char is lwr
					break;
				case 0xc8: // Latin Exteneded
					if ((*p >= 0x80)
						&& (*p <= 0x9f)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if ((*p >= 0xa2)
						&& (*p <= 0xb3)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if (*p == 0xbb)
						(*p)++; // Next char is lwr
					break;
				case 0xcd: // Greek & Coptic
					switch (*p) {
					case 0xb0:
					case 0xb2:
					case 0xb6:
						(*p)++; // Next char is lwr
						break;
					default:
						if (*p == 0xbf) {
							*pExtChar = 0xcf;
							(*p) = 0xb3;
						}
						break;
					}
				case 0xce: // Greek & Coptic
					if (*p == 0x86)
						(*p) = 0xac;
					else if (*p == 0x88)
						(*p) = 0xad;
					else if (*p == 0x89)
						(*p) = 0xae;
					else if (*p == 0x8a)
						(*p) = 0xaf;
					else if (*p == 0x8c) {
						*pExtChar = 0xcf;
						(*p) = 0x8c;
					}
					else if (*p == 0x8e) {
						*pExtChar = 0xcf;
						(*p) = 0x8d;
					}
					else if (*p == 0x8f) {
						*pExtChar = 0xcf;
						(*p) = 0x8e;
					}
					else if ((*p >= 0x91)
						&& (*p <= 0x9f))
						(*p) += 0x20; // US ASCII shift
					else if ((*p >= 0xa0)
						&& (*p <= 0xab)
						&& (*p != 0xa2)) {
						*pExtChar = 0xcf;
						(*p) -= 0x20;
					}
					break;
				case 0xcf: // Greek & Coptic
					if (*p == 0x8f)
						(*p) = 0xb4;
					else if (*p == 0x91)
						(*p)++; // Next char is lwr
					else if ((*p >= 0x98)
						&& (*p <= 0xaf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if (*p == 0xb4)
						(*p) = 0x91;
					else if (*p == 0xb7)
						(*p)++; // Next char is lwr
					else if (*p == 0xb9)
						(*p) = 0xb2;
					else if (*p == 0xbb)
						(*p)++; // Next char is lwr
					else if (*p == 0xbd) {
						*pExtChar = 0xcd;
						(*p) = 0xbb;
					}
					else if (*p == 0xbe) {
						*pExtChar = 0xcd;
						(*p) = 0xbc;
					}
					else if (*p == 0xbf) {
						*pExtChar = 0xcd;
						(*p) = 0xbd;
					}
					break;
				case 0xd0: // Cyrillic
					if ((*p >= 0x80)
						&& (*p <= 0x8f)) {
						*pExtChar = 0xd1;
						(*p) += 0x10;
					}
					else if ((*p >= 0x90)
						&& (*p <= 0x9f))
						(*p) += 0x20; // US ASCII shift
					else if ((*p >= 0xa0)
						&& (*p <= 0xaf)) {
						*pExtChar = 0xd1;
						(*p) -= 0x20;
					}
					break;
				case 0xd1: // Cyrillic supplement
					if ((*p >= 0xa0)
						&& (*p <= 0xbf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					break;
				case 0xd2: // Cyrillic supplement
					if (*p == 0x80)
						(*p)++; // Next char is lwr
					else if ((*p >= 0x8a)
						&& (*p <= 0xbf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					break;
				case 0xd3: // Cyrillic supplement
					if ((*p >= 0x81)
						&& (*p <= 0x8e)
						&& (*p % 2)) // Odd
						(*p)++; // Next char is lwr
					else if ((*p >= 0x90)
						&& (*p <= 0xbf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					break;
				case 0xd4: // Cyrillic supplement & Armenian
					if ((*p >= 0x80)
						&& (*p <= 0xaf)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					else if ((*p >= 0xb1)
						&& (*p <= 0xbf)) {
						*pExtChar = 0xd5;
						(*p) -= 0x10;
					}
					break;
				case 0xd5: // Armenian
					if ((*p >= 0x80)
						&& (*p <= 0x96)
						&& (!(*p % 2))) // Even
						(*p)++; // Next char is lwr
					break;
				case 0xe1: // Three byte code
					pExtChar = p;
					p++;
					switch (*pExtChar) {
					case 0x82: // Georgian
						if ((*p >= 0xa0)
							&& (*p <= 0xbf)) {
							*pExtChar = 0x83;
							(*p) -= 0x10;
						}
						break;
					case 0x83: // Georgian
						if ((*p >= 0x80)
							&& (((*p <= 0x85)
								|| (*p == 0x87))
							|| (*p == 0x8d)))
							(*p) += 0x30;
						break;
					case 0xb8: // Latin extened
						if ((*p >= 0x80)
							&& (*p <= 0xbf)
							&& (!(*p % 2))) // Even
							(*p)++; // Next char is lwr
						break;
					case 0xb9: // Latin extened
						if ((*p >= 0x80)
							&& (*p <= 0xbf)
							&& (!(*p % 2))) // Even
							(*p)++; // Next char is lwr
						break;
					case 0xba: // Latin extened
						if ((*p >= 0x80)
							&& (*p <= 0x94)
							&& (!(*p % 2))) // Even
							(*p)++; // Next char is lwr
						else if ((*p >= 0x9e)
							&& (*p <= 0xbf)
							&& (!(*p % 2))) // Even
							(*p)++; // Next char is lwr
						break;
					case 0xbb: // Latin extened
						if ((*p >= 0x80)
							&& (*p <= 0xbf)
							&& (!(*p % 2))) // Even
							(*p)++; // Next char is lwr
						break;
					case 0xbc: // Greek extened
						if ((*p >= 0x88)
							&& (*p <= 0x8f))
							(*p) -= 0x08;
						else if ((*p >= 0x98)
							&& (*p <= 0x9f))
							(*p) -= 0x08;
						else if ((*p >= 0xa8)
							&& (*p <= 0xaf))
							(*p) -= 0x08;
						else if ((*p >= 0xb8)
							&& (*p <= 0x8f))
							(*p) -= 0x08;
						break;
					case 0xbd: // Greek extened
						if ((*p >= 0x88)
							&& (*p <= 0x8d))
							(*p) -= 0x08;
						else if ((*p >= 0x98)
							&& (*p <= 0x9f))
							(*p) -= 0x08;
						else if ((*p >= 0xa8)
							&& (*p <= 0xaf))
							(*p) -= 0x08;
						else if ((*p >= 0xb8)
							&& (*p <= 0x8f))
							(*p) -= 0x08;
						break;
					case 0xbe: // Greek extened
						if ((*p >= 0x88)
							&& (*p <= 0x8f))
							(*p) -= 0x08;
						else if ((*p >= 0x98)
							&& (*p <= 0x9f))
							(*p) -= 0x08;
						else if ((*p >= 0xa8)
							&& (*p <= 0xaf))
							(*p) -= 0x08;
						else if ((*p >= 0xb8)
							&& (*p <= 0xb9))
							(*p) -= 0x08;
						break;
					case 0xbf: // Greek extened
						if ((*p >= 0x88)
							&& (*p <= 0x8c))
							(*p) -= 0x08;
						else if ((*p >= 0x98)
							&& (*p <= 0x9b))
							(*p) -= 0x08;
						else if ((*p >= 0xa8)
							&& (*p <= 0xac))
							(*p) -= 0x08;
						break;
					default:
						break;
					}
					break;
				case 0xf0: // Four byte code
					pExtChar = p;
					p++;
					switch (*pExtChar) {
					case 0x90:
						pExtChar = p;
						p++;
						switch (*pExtChar) {
						case 0x92: // Osage 
							if ((*p >= 0xb0)
								&& (*p <= 0xbf)) {
								*pExtChar = 0x93;
								(*p) -= 0x18;
							}
							break;
						case 0x93: // Osage 
							if ((*p >= 0x80)
								&& (*p <= 0x93))
								(*p) += 0x18;
							break;
						default:
							break;
						}
						break;
					default:
						break;
					}
					break;
				default:
					break;
				}
				pExtChar = 0;
			}
			p++;
		}
	}
	return pString;
}

/********************************************************************

int StrnCiCmp(const char *s1, const char *s2, size_t ztCount)

*********************************************************************/
int StrnCiCmp(const char *s1, const char *s2, size_t ztCount)
{
	if (s1 && *s1 && s2 && *s2) {
		unsigned char *pStr1Low = calloc(strlen(s1) + 1, sizeof(char));
		unsigned char *pStr2Low = calloc(strlen(s2) + 1, sizeof(char));
		if (pStr1Low && pStr2Low) {
			unsigned char *p1 = pStr1Low;
			unsigned char *p2 = pStr2Low;
			strcpy((char *)pStr1Low, s1);
			strcpy((char *)pStr2Low, s2);
			StrLwrExt(pStr1Low);
			StrLwrExt(pStr2Low);
			for (; ztCount--; p1++, p2++) {
				int iDiff = *p1 - *p2;
				if (iDiff != 0 || !*p1 || !*p2) {
					free(pStr1Low);
					free(pStr2Low);
					return iDiff;
				}
			}
			free(pStr1Low);
			free(pStr2Low);
			return 0;
		}
		return (-1);
	}
	return (-1);
}

/********************************************************************

int StrCiCmp(const char *s1, const char *s2)

*********************************************************************/
int StrCiCmp(const char *s1, const char *s2)
{
	return StrnCiCmp(s1, s2, (size_t)(-1));
}

/********************************************************************

char *StrCiStr(const char *s1, const char *s2)

*********************************************************************/
char *StrCiStr(const char *s1, const char *s2)
{
	if (s1 && *s1 && s2 && *s2) {
		char *p = (char *)s1;
		size_t ztWStringLength = strlen(s2);
		while (*p) {
			if (StrnCiCmp(p, s2, ztWStringLength) == 0)
				return p;
			p++;
		}
	}
	return (0);
}
#if 0
/********************************************************************

unsigned char *StrToLwrUft8Latin1(unsigned char *pString)

*********************************************************************/
unsigned char *StrToLwrUft8Latin1(unsigned char *pString)
{
	char cExtChar = 0;
	if (pString && *pString) {
		unsigned char *p = pString;
		while (*p) {
			if (((cExtChar && ((*p >= 0x80) && (*p <= 0xbf)))
				|| ((!cExtChar) && (*p <= 0x7f)))
				&& ((((*p & 0x7f) + cExtChar) >= 0x40)
					&& (((*p & 0x7f) + cExtChar) <= 0x5f)))
				(*p) += 0x20;
			if (cExtChar)
				cExtChar = 0;
			else if (*p == 0xc3)
				cExtChar = 0x40;
			p++;
		}
	}
	return pString;
}

/********************************************************************

unsigned char *StrToUprUft8Latin1(unsigned char *pString)

*********************************************************************/
unsigned char *StrToUprUft8Latin1(unsigned char *pString)
{
	char cExtChar = 0;
	if (pString && *pString) {
		unsigned char *p = pString;
		while (*p) {
			if (((cExtChar && ((*p >= 0x80) && (*p <= 0xbf)))
				|| ((!cExtChar) && (*p <= 0x7f)))
				&& ((((*p & 0x7f) + cExtChar) >= 0x60)
					&& (((*p & 0x7f) + cExtChar) <= 0x7e)))
				(*p) -= 0x20;
			if (cExtChar)
				cExtChar = 0;
			else if (*p == 0xc3)
				cExtChar = 0x40;
			p++;
		}
	}
	return pString;
}

/********************************************************************

int StrnCiCmpLatin1(const char *s1, const char *s2, size_t ztCount)

*********************************************************************/
int StrnCiCmpLatin1(const char *s1, const char *s2, size_t ztCount)
{
	unsigned char cExtChar = 0;
	if (s1 && *s1 && s2 && *s2) {
		for (; ztCount--; s1++, s2++) {
			int iDiff = tolower((unsigned char)(*s1 & 0x7f)
				+ cExtChar) - tolower((unsigned char)(*s2 & 0x7f) + cExtChar);
			if (iDiff != 0 || !*s1 || !*s2)
				return iDiff;
			if (cExtChar)
				cExtChar = 0;
			else if (((unsigned char )*s2) == ((unsigned char)0xc3))
				cExtChar = 0x40;
		}
	}
	return 0;
}

/********************************************************************

unsigned char *StrToUprUft8Latin(unsigned char *pString)

*********************************************************************/
int StrCiCmpLatin1(const char *s1, const char *s2)
{
	return StrnCiCmpLatin1(s1, s2, (size_t)(-1));
}

/********************************************************************

char *StrCiStrLatin1(const char *s1, const char *s2)

*********************************************************************/
char *StrCiStrLatin1(const char *s1, const char *s2)
{
	if (s1 && *s1 && s2 && *s2) {
		char *p = (char *)s1;
		size_t ztWStringLength = strlen(s2);
		while (*p) {
			if (StrnCiCmpLatin1(p, s2, ztWStringLength) == 0)
				return p;
			p++;
		}
	}
	return (0);
}
#endif
/********************************************************************

String functions

*********************************************************************/
/*************************************************************

char *StringUpdate(char **ppOriginalData, char *pNewData)

*************************************************************/
char *StringUpdate(char **ppOriginalData, char *pNewData)
{
	if(pNewData && (!*ppOriginalData || (ppOriginalData && *ppOriginalData && strcmp(*ppOriginalData, pNewData))))
		CopyString(ppOriginalData, pNewData);
	if(ppOriginalData && (*ppOriginalData))
		{
		char *pOriginalData = *ppOriginalData;
		return pOriginalData;
		}
	else
		return NulString;
}

/*************************************************************

char *CopyString(char **ppTarget, char *pData)

*************************************************************/
char *CopyString(char **ppTarget, const char *pData)
{
	if (pData)
		{
		char *pCopy = 0;
		size_t ztLen = 0;
		if (pData)
			ztLen = strlen(pData);
		if(ppTarget)
			{
			// If Data the same, no change
			if((*ppTarget) && (!strcmp(*ppTarget, pData)))
				return *ppTarget;
			FreeString(ppTarget);
			}
		pCopy = calloc(ztLen + 1, sizeof(char));
		if (pCopy)
			{
			strcpy(pCopy, pData);
			// Wipe ending spaces
			while (*pCopy && (pCopy[strlen(pCopy) - 1] == ' '))
				pCopy[strlen(pCopy) - 1] = 0;
			}
		if (ppTarget)
			*ppTarget = pCopy;
		return pCopy;
		}
	return NulString;
}

/*************************************************************

char *FreeString(char **ppData)

*************************************************************/
char *FreeString(char **ppData)
{
	if (*ppData && (*ppData != NulString))
		free(*ppData);
	*ppData = NulString;
	return *ppData;
}

/********************************************************************

void ReadString(char **ppData, char *pRead)

*********************************************************************/
void ReadString(char **ppData, char *pRead)
{
	CopyString(ppData, pRead);
	free(pRead);
}

/********************************************************************

File functions

*********************************************************************/
/*************************************************************

FILE *FileOpen(char *pFileName, char *pMode)

*************************************************************/
FILE *FileOpen(char *pFileName, char *pMode)
{
#ifdef UNICODE
	FILE *Utf8WinFileOpen(char *pFileName, char *pMode);
	return Utf8WinFileOpen(pFileName, pMode);
#else // !UNICODE
	return fopen((const char *)pFileName, (const char *)pMode);
#endif
}

/*************************************************************

char *FileRead(char *szFileName)

*************************************************************/
char *FileRead(char *szFileName)
{
	// Open the file on disk
	FILE *pFile = FileOpen(szFileName, "rb");
	if (!pFile)
		return 0; // File not found or opened
	{
	size_t			ztMemory = 0x10000;
	char			*pBuffer = (char *)calloc(ztMemory, sizeof(char));
	if (pBuffer)
		{
		char	*pTmp = 0;
		size_t	ztRead = fread(
			pBuffer,
			sizeof(char),
			ztMemory,
			pFile);
		while(ztMemory <= ztRead)
			{
			ztMemory *= 2;
			pTmp = (char *)realloc(pBuffer, ztMemory);
			if(pTmp)
				pBuffer = pTmp;
			else
				{
				free(pBuffer);
				return 0;
				}
                memset(pBuffer + ztRead, 0, ztMemory - ztRead);
			ztRead += fread(
				pBuffer + ztRead,
				sizeof(char),
				ztMemory - ztRead,
				pFile);
			}
		fclose(pFile);
		pTmp = (char *)realloc(pBuffer, ztRead + 1);
		if(pTmp)
			pBuffer = pTmp;
		else
			{
			free(pBuffer);
			return 0;
			}
		}
	return(pBuffer);
	}
}

/*************************************************************

size_t FileWrite(char *szFileName, char *pBuffer)

*************************************************************/
size_t FileWrite(char *szFileName, char *pBuffer)
{
	// Open the file on disk
	FILE *pFile = FileOpen(szFileName, "wb");
	if (!pFile)
		return 0; // File not found or opened
	{
	size_t ztWritten = fwrite(pBuffer, sizeof(char), strlen(pBuffer), pFile);
	fclose(pFile);
	return ztWritten;
	}
}

/********************************************************************

Xml Syntax functions

*********************************************************************/
/*************************************************************

size_t ufWrite(FILE *pStream, char *pBuffer)

*************************************************************/
int ufWrite(FILE *pStream, char *pBuffer)
{
	if (!pBuffer)
		return 0;

	if (!pStream)
		return (-2);

	{
		size_t ztBytesWritten = fwrite(
			pBuffer,
			sizeof(char),
			strlen(pBuffer),
			pStream);

		if (strlen(pBuffer) != ztBytesWritten)
			return (-2);

		return (int)ztBytesWritten;
	}
}


/*************************************************************

size_t xmlWrite(FILE *pStream, char *pBuffer)

*************************************************************/
size_t xmlWrite(FILE *pStream, char *pBuffer)
{
	// This is only for raw text strings that can't be XML commands, and for such ESC-seqences are necessary for &<>' and "
	size_t ztBufferLen = strlen(pBuffer);
	size_t ztSentChar = 0;

	if (!pBuffer)
		return 0;

	if (!pStream)
		return (size_t)(-2);

	if (ztBufferLen)
		{
		unsigned char *pSendBuffer = calloc(ztBufferLen + 1, sizeof(char));
		if (pSendBuffer)
			{
			unsigned char *pSendPtr = pSendBuffer;
			unsigned char *pTestPtr = pSendPtr;
			strcpy((char *)pSendBuffer, pBuffer);

			while (*pTestPtr)
				{
				switch (*pTestPtr)
					{
					case '&':
						*pTestPtr = 0;
						ztSentChar += ufWrite(pStream, (char *)pSendPtr);
						ztSentChar += ufWrite(pStream, "&amp;");
						pSendPtr = (pTestPtr + 1);
						*pTestPtr = '&';
						break;
					case '<':
						*pTestPtr = 0;
						ztSentChar += ufWrite(pStream, (char *)pSendPtr);
						ztSentChar += ufWrite(pStream, "&lt;");
						pSendPtr = (pTestPtr + 1);
						*pTestPtr = '<';
						break;
					case '>':
						*pTestPtr = 0;
						ztSentChar += ufWrite(pStream, (char *)pSendPtr);
						ztSentChar += ufWrite(pStream, "&gt;");
						pSendPtr = (pTestPtr + 1);
						*pTestPtr = '>';
						break;
/*					case '\'':
						*pTestPtr = 0;
						ztSentChar += ufWrite(pStream, (char *)pSendPtr);
						ztSentChar += ufWrite(pStream, "&apos;");
						pSendPtr = (pTestPtr + 1);
						*pTestPtr = '\'';
						break;
*/					case '"':
						*pTestPtr = 0;
						ztSentChar += ufWrite(pStream, (char *)pSendPtr);
						ztSentChar += ufWrite(pStream, "&quot;");
						pSendPtr = (pTestPtr + 1);
						*pTestPtr = '"';
						break;
					default:
						if (*pTestPtr < 0x20) // No ctrl chars
							{
							*pTestPtr = 0;
							ztSentChar += ufWrite(pStream, (char *)pSendPtr);
							ztSentChar += ufWrite(pStream, " ");
							pSendPtr = (pTestPtr + 1);
							*pTestPtr = '"';
							}
						break;
					}
				pTestPtr++;
				}
			if (*pSendPtr)
				ztSentChar += ufWrite(pStream, (char *)pSendPtr);

			free(pSendBuffer);
			}
		}
	else
		ztBufferLen = ztBufferLen;
	return ztSentChar;
}


/********************************************************************
*
int xmlWriteLine(FILE *pStream, int iTab, char *attr, char *pPrefix, char *pData, char *pSuffix)
*
*********************************************************************/
int xmlWriteLine(FILE *pStream, int iTab, char *attr, char *pPrefix, char *pData, char *pSuffix)
{
	static char LineBuffer[WRITE_BUFFER_LENGTH];
	size_t ztBytes = 0;
	if (iTab<0)
		return FALSE;

	memset(LineBuffer, 0, WRITE_BUFFER_LENGTH);
	// Tabs
	while (iTab--)
		strcat(LineBuffer, "\t");
	if (pPrefix)
	{
		strcat(LineBuffer, "<");
		if (pPrefix[0])
			strcat(LineBuffer, pPrefix);
		// ContextRef
		if (attr && *attr)
		{
			strcat(LineBuffer, " ");
			ztBytes += ufWrite(pStream, LineBuffer);
			memset(LineBuffer, 0, WRITE_BUFFER_LENGTH);

			ztBytes += ufWrite(pStream, attr);
			memset(LineBuffer, 0, WRITE_BUFFER_LENGTH);
		}
		// pData
		if (((!pData) || (!(*pData))) && pPrefix && pSuffix)
		{
			strcat(LineBuffer, " xsi:nil=\"true\"/");
			pSuffix = 0;
		}
		strcat(LineBuffer, ">");
		ztBytes += ufWrite(pStream, LineBuffer);
		memset(LineBuffer, 0, WRITE_BUFFER_LENGTH);

		if (pData && *pData)
			ztBytes += xmlWrite(pStream, pData);
	}
	if (pSuffix)
	{
		strcat(LineBuffer, "</");
		strcat(LineBuffer, pSuffix);
		strcat(LineBuffer, ">");
		ztBytes += ufWrite(pStream, LineBuffer);
		memset(LineBuffer, 0, WRITE_BUFFER_LENGTH);
	}
	// Newline
	ztBytes += ufWrite(pStream, "\r\n");
	return TRUE;
}

/********************************************************************

XmlTagDataType *GetNextXmlTag(XmlTagDataType *XmlTagData)

*********************************************************************/
XmlTagDataType *GetNextXmlTag(XmlTagDataType *XmlTagData)
{
#define TAG_BEGIN	'<'
#define TAG_END		'>'
#define TAG_NAME	':'

	BOOL bNoData = FALSE;
	BOOL bEnd = FALSE;

	if (!XmlTagData)
		return 0;

	{
		char cTagNameTag = ':';

		char *pTagStart = strchr(XmlTagData->pBuffer, TAG_BEGIN);
		char *pTagEnd = 0;
		char *pNextTagStart = 0;
		char *pNextTagEnd = 0;

		if (!pTagStart)
			return 0;
		pTagEnd = strchr(pTagStart, TAG_END);
		if (!pTagEnd)
			return 0;

		// Non-data stuff
		while (((pTagStart)
			&& (pTagEnd))
			&& ((pTagStart[1] == '?')
				|| (pTagStart[1] == '!')
				|| (pTagEnd[-1] == '/')
				|| (pTagEnd[-1] == '-')
				|| (pTagEnd[-1] == '?')))
		{
			pTagStart = strchr(pTagEnd, TAG_BEGIN);
			pTagEnd = strchr(pTagStart, TAG_END);
		}
		pNextTagStart = strchr(pTagEnd, TAG_BEGIN);
		if (pNextTagStart)
			pNextTagEnd = strchr(pNextTagStart, TAG_END);
		// else pNextTagEnd=0; End of the file

		// Non-data stuff
		while (((pNextTagStart)
			&& (pNextTagEnd))
			&& ((pNextTagStart[1] == '?')
				|| (pNextTagStart[1] == '!')
				|| (pNextTagEnd[-1] == '/')
				|| (pNextTagEnd[-1] == '-')
				|| (pNextTagEnd[-1] == '?')))
		{
			if (pNextTagEnd[-1] == '/')
				bNoData = TRUE;
			pNextTagStart = strchr(pNextTagEnd, TAG_BEGIN);
			pNextTagEnd = strchr(pNextTagStart, TAG_END);
		}

		if (pTagStart && pTagEnd)
		{
			// Get the right endpoint of the first tag lable
			char *pTagAttrStart = pTagStart;
			char *pTagAttrEnd = pTagEnd;
			char *pTagDataStart = pTagEnd + 1;
			char *pTagDataEnd = pNextTagStart;
			if (!bNoData)
			{
				if (!pNextTagStart) // Final tags do not have data
					pTagDataEnd = pTagDataStart;
			}
			// Find the real end the tag (besides attributes and such)
			pTagEnd = pTagStart;
			while ((' '<*pTagEnd) && (pTagEnd<pTagAttrEnd))
				pTagEnd++;
			// Get the attribute of the first tag lable
			pTagAttrStart = pTagEnd;
			if (pTagAttrStart<pTagAttrEnd) // Any attribute
				pTagAttrStart++;
			while ((' ' >= *pTagAttrStart) && (pTagAttrStart<pTagAttrEnd))
				pTagAttrStart++;
			if (!bNoData)
			{
				// Find the real Data start to get ride of spaces and tabs etc
				while ((' ' >= *pTagDataStart) && (0<*pTagDataStart) && (pTagDataStart<pTagDataEnd))
					pTagDataStart++;
			}
			// To syncron one step from the text (if right </... to /...
			if (pNextTagStart)
				pNextTagStart++;
			// Termination for char search
			*pTagEnd = 0;
			if (pNextTagStart)
			{
				*pTagAttrEnd = 0;
				*pTagDataEnd = 0;
			}
			if (pNextTagEnd)
				*pNextTagEnd = 0;
			// Remove the prenames (before any :), we just don't care of prenames
			if (pTagStart[1] == '/')
				bEnd = TRUE;
			if (strchr(pTagStart, cTagNameTag))
				pTagStart = strchr(pTagStart, cTagNameTag);
			if (pNextTagStart && strchr(pNextTagStart, cTagNameTag))
				pNextTagStart = strchr(pNextTagStart, cTagNameTag);
			// Step one ahead to point at the tag lables
			pTagStart++;
			if (pNextTagStart)
				pNextTagStart++;
			// Compare if a pair
			if ((!bEnd) && pNextTagStart && (!strcmp(pTagStart, pNextTagStart)))	// Data pair
			{
				// Pair = Data
				*pNextTagEnd = TAG_END;
				XmlTagData->pTag = pTagStart;
				XmlTagData->Attr = pTagAttrStart;
				if (!bNoData)
					XmlTagData->pData = pTagDataStart;
				else
					XmlTagData->pData = 0;
				if ((!bNoData) && XmlTagData->pData && *XmlTagData->pData)
					XmlTagData->pBuffer = (pNextTagEnd + 1);
				else
				{ // A blank pair don't count, most likly a blank mandatory command or junk
					if (pTagAttrEnd != pTagEnd)
						*pTagAttrEnd = TAG_END;
					if (pTagDataEnd)
						*pTagDataEnd = TAG_BEGIN;
					if (pNextTagEnd)
						*pNextTagEnd = TAG_END;
					XmlTagData->pTag = pTagStart;
					XmlTagData->Attr = 0;
					XmlTagData->pData = 0;
					//				XmlTagData->pBuffer=pTagEnd+1;
					XmlTagData->pBuffer = (pNextTagEnd + 1);
				}
			}
			else														// Down or Up
			{
				// No pair = Command
				if ((pTagAttrEnd != pTagEnd) && pNextTagStart)
					*pTagAttrEnd = TAG_END;
				if (pTagDataEnd && pNextTagStart)
					*pTagDataEnd = TAG_BEGIN;
				if (pNextTagEnd)
					*pNextTagEnd = TAG_END;
				if (pTagStart && (*pTagStart == '/'))
					pTagStart++;
				XmlTagData->pTag = pTagStart;
				XmlTagData->Attr = 0;
				XmlTagData->pData = 0;
				XmlTagData->pBuffer = pTagEnd + 1;
			}
		}
	}

	return XmlTagData;
}

/********************************************************************

char *CloseXmlAttrString(char *pText, char *pTag)

*********************************************************************/
char *CloseXmlAttrString(char *pText, char *pTag)
{
	char *pAttrData =strstr(pText, pTag);
	if (pAttrData)
		{
		pAttrData = strchr(pAttrData, '"');
		if (pAttrData)
			{
			char *pEnd = strchr(++pAttrData, '"');
			*pEnd = 0;
			return pAttrData;
			}
		}
	return 0;
}


/*************************************************************

char *ClearXmlEsc(char *pBuffer)

*************************************************************/
char *ClearXmlEsc(char *pBuffer)
{
	// This is only for raw text strings that can't be XML commands, and for such ESC-seqences are necessary for &<>' and "
	size_t BufferLen = strlen(pBuffer);

	// &
	char *pPtr = strstr(pBuffer, "&amp;");
	while (pPtr)
	{
		size_t iEscString = strlen("&amp;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '&';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&amp;");
	}
	// <
	pPtr = strstr(pBuffer, "&lt;");
	while (pPtr)
	{
		size_t iEscString = strlen("&lt;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '<';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&lt;");
	}
	// >
	pPtr = strstr(pBuffer, "&gt;");
	while (pPtr)
	{
		size_t iEscString = strlen("&gt;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '>';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&gt;");
	}
	// <=
	pPtr = strstr(pBuffer, "&le;");
	while (pPtr)
		{
		size_t iEscString = strlen("&lt;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr+1, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '<';
		*(pPtr + 1) = '=';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&le;");
		}
	// >=
	pPtr = strstr(pBuffer, "&ge;");
	while (pPtr)
		{
		size_t iEscString = strlen("&gt;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr + 1, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '>';
		*(pPtr+1) = '=';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&ge;");
		}
	// "
	pPtr = strstr(pBuffer, "&quot;");
	while (pPtr)
	{
		size_t iEscString = strlen("&quot;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '\"';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&quot;");
	}
	// '
	pPtr = strstr(pBuffer, "&apos;");
	while (pPtr)
	{
		size_t iEscString = strlen("&apos;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '\'';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&apos;");
	}
	// '
	pPtr = strstr(pBuffer, "&#39;");
	while (pPtr)
	{
		size_t iEscString = strlen("&#39;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '\'';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&#39;");
	}
	// '
	pPtr = strstr(pBuffer, "&#x0027;");
	while (pPtr)
	{
		size_t iEscString = strlen("&#x0027;");
		BufferLen -= (iEscString - 1);
		memmove(pPtr, (pPtr + (iEscString - 1)), BufferLen - (pPtr - pBuffer));
		*pPtr = '\'';
		pBuffer[BufferLen] = 0;
		pPtr = strstr(pPtr, "&#x0027;");
	}

	return pBuffer;
}

/********************************************************************

char *ReadXmlDataString(char *Indata)

*********************************************************************/
char *ReadXmlDataString(char *Indata)
{
	if (!Indata)
		return 0;
	{
		char *Data = calloc((strlen(Indata) * 2) + 1, sizeof(char));
		if (Data)
			{
			strcpy(Data, Indata);
			Data = ClearXmlEsc(Data);
			}
		return Data;
	}
}

/*********************************************************************

Linked list functions

The linked list is based on the struct LLISTITEMTYPE that contains a void pointer to the Data
and reference to the Next Record and the PreviousRecord.

Three of these forms the LLISTTYPE set of the First Record, the Cursor Record and the Last Record.
This way there are entrances and ability to wind the linked list back and forward to retrieve data.

The functions in LList.c are to handle the linked list database system.

************************************************************************/
/***********************************************************************
LLIST LListInit(void)

This function created a linked list header
************************************************************************/
LLIST LListInit(void)
{
	// Alloc pLList data structure
	LLIST pLList = (LLIST)calloc(1, sizeof(LLISTTYPE));
	if (!(pLList))
		return RET_ERR; // Failure

	memset(pLList, 0, sizeof(LLISTTYPE));
	return pLList; // Success
}


/***********************************************************************
BOOL LListDeleteData(LLIST pLList)

This function destroyes the linked list with data, records and keep the header
************************************************************************/
BOOL LListDeleteData(LLIST pLList)
{
	// Free record memory from the end
	if (!pLList)
		return RET_ERR; // pLList did not exist
	while (pLList->FirstRecord) { // As long as list exists
		LLISTITEMTYPE *pNextRecord = (LLISTITEMTYPE *)pLList->FirstRecord->vpNextRecord;
		if (pLList->FirstRecord->vpData)
			free(pLList->FirstRecord->vpData);  // Free Data area
		free(pLList->FirstRecord);			// Free pLList area
		pLList->FirstRecord = pNextRecord;
	}
	memset(pLList, 0, sizeof(LLISTTYPE));
	return RET_OK; // OK
}

/***********************************************************************
BOOL LListDestroy(LLIST pLList)

This function destroyes the linked list with data, records and header
************************************************************************/
BOOL LListDestroy(LLIST pLList)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	LListDeleteData(pLList);
	if (pLList)
		free(pLList);			// Free pLList area
	pLList = 0;
	return RET_OK; // OK
}

/***********************************************************************
BOOL LListClear(LLIST pLList)

This function destroyes the linked list records BUT keeps the data and header

************************************************************************/
BOOL LListClear(LLIST pLList)
{
	// Free record memory from the end
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (pLList->FirstRecord) {			// pLList exists but with records
		while (pLList->FirstRecord) {	// As long as list exists
			LLISTITEMTYPE *pNextRecord = (LLISTITEMTYPE *)pLList->FirstRecord->vpNextRecord;
			free(pLList->FirstRecord);			// Free pLList area
			pLList->FirstRecord = pNextRecord;
		}
		// All records are now deleted
	}

	memset(pLList, 0, sizeof(LLISTTYPE));
	return RET_OK; // OK
}


/***********************************************************************
BOOL LListErase(LLIST pLList)

This function destroyes the linked list, records and header BUT keep the data
(This is used when data pointers are stored in list and combo boxes
************************************************************************/
BOOL LListErase(LLIST pLList)
{
	LListClear(pLList);
	if (pLList)
		free(pLList);			// Free pLList area
	pLList = 0;
	return RET_OK; // OK
}




/***********************************************************************
BOOL LListAddRecord(LLIST pLList, void *vpData)

This function adds a record with data to the linked list
************************************************************************/
BOOL LListAddRecord(LLIST pLList, void *vpData)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist
//	if (!vpData)
//		return RET_ERR; 

	{
		LLISTITEMTYPE *NewRecord = (LLISTITEMTYPE *)calloc(1, sizeof(LLISTITEMTYPE));
		if (!NewRecord)
			return RET_ERR; // No data area allocated
		memset(NewRecord, 0, sizeof(LLISTITEMTYPE));

		NewRecord->vpData = vpData;

		if (pLList->FirstRecord) {
			NewRecord->vpPrevRecord = pLList->LastRecord;
			pLList->LastRecord->vpNextRecord = NewRecord;
		}
		else { // No records in the list
			pLList->FirstRecord = NewRecord;
		}
		pLList->LastRecord = NewRecord;
		pLList->CursorRecord = NewRecord;
	}
	return RET_OK;
}

/***********************************************************************
BOOL LListUpdateDataCursorRecord(LLIST pLList, void *vpNewData)

This function updates the data of the cursorRecord
************************************************************************/
BOOL LListUpdateDataCursorRecord(LLIST pLList, void *vpNewData)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (pLList->CursorRecord) {
		pLList->CursorRecord->vpData = vpNewData;
		return RET_OK;
	}
	return RET_ERR; // Error
}


/***********************************************************************
BOOL LListSearchDeleteData(LLIST pLList, void *vpSearchData)

This function finds the record with the same data pointer and deteltes the record
************************************************************************/
void *LListSearchDeleteRecord(LLIST pLList, void *vpSearchData)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist
	if (pLList->FirstRecord) {						// pLList exists but with records
		pLList->CursorRecord = pLList->FirstRecord;
		while (pLList->CursorRecord->vpNextRecord) {	// As long as list exists
			if (pLList->CursorRecord->vpData == vpSearchData)
				return LListDeleteCursorRecord(pLList);

			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		}
		if (pLList->CursorRecord->vpData == vpSearchData)
			return LListDeleteCursorRecord(pLList);
	}
	return RET_ERR; // Error
}


/***********************************************************************
void **LListSearchData(LLIST pLList, void *vpSearchData)

This function finds the record with the same data pointer and returns the pointer to the datapointer
************************************************************************/
void **LListSearchData(LLIST pLList, void *vpSearchData)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (pLList->FirstRecord) {						// pLList exists but with records
		pLList->CursorRecord = pLList->FirstRecord;

		while (pLList->CursorRecord->vpNextRecord) {	// As long as list exists
			if (pLList->CursorRecord->vpData == vpSearchData)
				return &pLList->CursorRecord->vpData;
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		}
		if (pLList->CursorRecord->vpData == vpSearchData)
			return &pLList->CursorRecord->vpData;
	}
	return RET_ERR; // Error
}


/***********************************************************************
void **LListSearchText(LLIST pLList, char *szSearchText)

This function finds the record with the same char string and returns the pointer to the datapointer
************************************************************************/
void **LListSearchText(LLIST pLList, char *szSearchText)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (pLList->FirstRecord) {						// pLList exists but with records
		pLList->CursorRecord = pLList->FirstRecord;

		// Free record memory from the end
		while (pLList->CursorRecord->vpNextRecord) {	// As long as list exists
			char *pListText = pLList->CursorRecord->vpData;
			if ((!*pListText) && (*szSearchText))
				return &pLList->CursorRecord->vpData;
			if (!strcmp(pLList->CursorRecord->vpData, szSearchText))
				return &pLList->CursorRecord->vpData;
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		}
		if (!strcmp(pLList->CursorRecord->vpData, szSearchText))
			return &pLList->CursorRecord->vpData;
	}
	return RET_ERR; // Error
}


/***********************************************************************
void LListNulDataCursorRecord(LLIST pLList)

This function deletes a record with data from a linked list
************************************************************************/
void LListNulDataCursorRecord(LLIST pLList)
{
	pLList->CursorRecord->vpData = 0;
}


/***********************************************************************
void *LListDeleteCursorRecord(LLIST pLList)

This function deletes a record with data from a linked list
************************************************************************/
void *LListDeleteCursorRecord(LLIST pLList)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (!pLList->CursorRecord)  // pLList exists
		return RET_ERR; // pLList did not exist

	{
		// Fix up the linked list
		LLISTITEMTYPE *OldCursorRecord = pLList->CursorRecord;
		LLISTITEMTYPE *OldNextRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		LLISTITEMTYPE *OldPreviousRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
		if (OldPreviousRecord)
			OldPreviousRecord->vpNextRecord = pLList->CursorRecord->vpNextRecord;
		else {
			// We are deleteing the first record
			pLList->FirstRecord = pLList->FirstRecord->vpNextRecord;
			if (pLList->FirstRecord) // If there are any more records
				pLList->FirstRecord->vpPrevRecord = 0;
		}
		if (OldNextRecord)
			OldNextRecord->vpPrevRecord = pLList->CursorRecord->vpPrevRecord;
		else if (pLList->FirstRecord) { // If there are any more records
			// We are deleteing the last record
			pLList->LastRecord = pLList->CursorRecord->vpPrevRecord;
			if (pLList->LastRecord) // If there are any more records
				pLList->LastRecord->vpNextRecord = 0;
		}
		else // There are no more records
			pLList->LastRecord = 0; // No record left

		if (pLList->CursorRecord->vpPrevRecord)
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
		else if (pLList->CursorRecord->vpNextRecord)
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		else
			pLList->CursorRecord = 0; // No record left

		// Free data
		if (OldCursorRecord->vpData)
			free(OldCursorRecord->vpData);
		free(OldCursorRecord);
	}
	if (pLList->CursorRecord)
		return pLList->CursorRecord->vpData;
	else
		return 0;
}

/***********************************************************************
void *LListEraseCursorRecord(LLIST pLList)

This function erases a record but not the data from a linked list
************************************************************************/
void *LListEraseCursorRecord(LLIST pLList)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (!pLList->CursorRecord)  // pLList exists
		return RET_ERR; // pLList did not exist

	{
		// Fix up the linked list
		LLISTITEMTYPE *OldCursorRecord = pLList->CursorRecord;
		LLISTITEMTYPE *OldNextRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		LLISTITEMTYPE *OldPreviousRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
		if (OldPreviousRecord)
			OldPreviousRecord->vpNextRecord = pLList->CursorRecord->vpNextRecord;
		else { // We are deleteing the first record
			pLList->FirstRecord = pLList->FirstRecord->vpNextRecord;
			if (pLList->FirstRecord) // If there are any more records
				pLList->FirstRecord->vpPrevRecord = 0;
		}
		if (OldNextRecord)
			OldNextRecord->vpPrevRecord = pLList->CursorRecord->vpPrevRecord;
		else if (pLList->FirstRecord) { // If there are any more records
			// We are deleteing the last record
			pLList->LastRecord = pLList->CursorRecord->vpPrevRecord;
			if (pLList->LastRecord) // If there are any more records
				pLList->LastRecord->vpNextRecord = 0;
		}
		else // There are no more records
			pLList->LastRecord = 0; // No record left

		if (pLList->CursorRecord->vpPrevRecord)
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
		else if (pLList->CursorRecord->vpNextRecord)
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
		else
			pLList->CursorRecord = 0; // No record left

		free(OldCursorRecord);
	}
	if (pLList->CursorRecord)
		return pLList->CursorRecord->vpData;
	else
		return 0;
}

/***********************************************************************
void *LListGetCursorRecordData(LLIST pLList)

This function get the void data from a linked list by pLList Cursor
************************************************************************/
void *LListGetCursorRecordData(LLIST pLList)
{
	if (pLList && pLList->LastRecord) { // pLList exists
		if (pLList->CursorRecord) // Record found
			return pLList->CursorRecord->vpData;
	}
	return 0; // Data not found
}

/***********************************************************************
void *LListGetFirstRecordData(LLIST pLList)

This function get the void data from first record in the linked list
************************************************************************/
void *LListGetFirstRecordData(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {	// pLList exists
		if (pLList->FirstRecord) {		// Record found
			pLList->CursorRecord = pLList->FirstRecord;
			return pLList->CursorRecord->vpData;
		}
	}
	return 0; // Data not found
}

/***********************************************************************
void *LListGetNextRecordData(LLIST pLList)

This function get the void data from next record in the linked list
************************************************************************/
void *LListGetNextRecordData(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {				// pLList exists
		if (pLList->CursorRecord->vpNextRecord) {	// Record found
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
			return pLList->CursorRecord->vpData;
		}
	}
	return 0; // Data not found
}

/***********************************************************************
void *LListGetLastRecordData(LLIST pLList)

This function get the void data from last record in the linked list
************************************************************************/
void *LListGetLastRecordData(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {		// pLList exists
		if (pLList && pLList->LastRecord) {	// Record found
			pLList->CursorRecord = pLList->LastRecord;
			return pLList->CursorRecord->vpData;
		}
	}
	return 0; // Data not found
}

/***********************************************************************
void *LListGetPreviousRecordData(LLIST pLList)

This function get the void data from previous record in the linked list
************************************************************************/
void *LListGetPreviousRecordData(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {				// pLList exists
		if (pLList->CursorRecord->vpPrevRecord) {	// Record found
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
			return pLList->CursorRecord->vpData;
		}
	}
	return 0; // Data not found
}

/***********************************************************************
BOOL LListMoveCursorForward(LLIST pLList)

This function moves cursor forward
************************************************************************/
BOOL LListMoveCursorForward(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {				// pLList exists
		if (pLList->CursorRecord->vpNextRecord) {	// Record found
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpNextRecord;
			return RET_OK;
		}
	}
	return RET_ERR; // Data not found
}

/***********************************************************************
BOOL LListMoveCursorBackward(LLIST pLList)

This function moves cursor backward
************************************************************************/
BOOL LListMoveCursorBackward(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {				// pLList exists
		if (pLList->CursorRecord->vpPrevRecord) {	// Record found
			pLList->CursorRecord = (LLISTITEMTYPE *)pLList->CursorRecord->vpPrevRecord;
			return RET_OK;
		}
	}
	return RET_ERR; // Data not found
}

/***********************************************************************
BOOL LListMoveCursorToFirst(LLIST pLList)

This function moves cursor to the beginning of the list
************************************************************************/
BOOL LListMoveCursorToFirst(LLIST pLList)
{
	if (pLList->FirstRecord) {	// Record found
		pLList->CursorRecord = pLList->FirstRecord;
		return RET_OK;
	}
	return RET_ERR; // Data not found
}

/***********************************************************************
BOOL LListMoveCursorToLast(LLIST pLList)

This function moves cursor to the end of the list
************************************************************************/
BOOL LListMoveCursorToLast(LLIST pLList)
{
	if (pLList && pLList->LastRecord) {	// Record found
		pLList->CursorRecord = pLList->LastRecord;
		return RET_OK;
	}
	return RET_ERR; // Data not found
}


/***********************************************************************
LLIST LListCopyLList(LLIST *ppTargetLList,LLIST pOriginalLList)

Copy one list to another
************************************************************************/
LLIST LListCopyLList(LLIST *ppTargetLList, LLIST pOriginalLList)
{
	int iCounter = 0;

	if (*ppTargetLList)	// Init LList if not initiated
		LListClear(*ppTargetLList);
	else
		*ppTargetLList = LListInit();

	// Copy Data
	{
		// Read first record
		void *vpDataSource = LListGetFirstRecordData(pOriginalLList);
		void *vpDataTarget = 0;
		while (vpDataSource) {
			// Copy Data record pointer
			vpDataTarget = vpDataSource;
			// Add it to Target list
			LListAddRecord(*ppTargetLList, vpDataTarget);
			// Get next record from Original
			vpDataSource = LListGetNextRecordData(pOriginalLList);
			iCounter++;
		}
	}
	return *ppTargetLList;
}


/***********************************************************************
LLIST LListCopyLListWithContent(LLIST pTargetLList,LLIST pOriginalLList, int DataSize)

Copy one list to another
************************************************************************/
LLIST LListCopyLListWithContent(LLIST pTargetLList, LLIST pOriginalLList, int DataSize)
{
	int iCounter = 0;
	if (!DataSize)
		return 0;
	if (pTargetLList)	// Init LList if not initiated
		LListClear(pTargetLList);
	else
		return pTargetLList; // Must be initialised

	// Copy Data
	{
		// Read first record
		void *vpDataSource = LListGetFirstRecordData(pOriginalLList);
		void *vpDataTarget = 0;
		while (vpDataSource) {
			// Make Data memory for new record
			vpDataTarget = calloc(DataSize, sizeof(char)); // We must create new data areas to really make a copy of its content
			// Copy Data record
			if (vpDataTarget) {
				memcpy(vpDataTarget, vpDataSource, DataSize);
				// Add it to Target list
				LListAddRecord(pTargetLList, vpDataTarget);
				// Get next record from Original
				vpDataSource = LListGetNextRecordData(pOriginalLList);
			}
			iCounter++;
		}
	}
	return pTargetLList;
}


/***********************************************************************
unsigned long long LListGetListSize(LLIST pLList)

Copy one list to another
************************************************************************/
unsigned long long LListGetListSize(LLIST pLList)
{
	unsigned long long ullCounter = 0;

	if (!(pLList && pLList->LastRecord))
		return 0; // pLList did not exist
	if (LListGetFirstRecordData(pLList))
		ullCounter++;
	while (LListGetNextRecordData(pLList))
		ullCounter++;
	return ullCounter;
}



/***********************************************************************
void *LListGetCursorRecordPtr(LLIST pLList)

This function updates the data of the cursorRecord
************************************************************************/
void *LListGetCursorRecordPtr(LLIST pLList)
{
	if (!pLList)
		return RET_ERR; // pLList did not exist

	if (pLList->CursorRecord)
		return &pLList->CursorRecord->vpData;
	return RET_ERR; // pLList did not exist
}

