/*********************************************************************************

	RosettaFioUtils.h

	Alphabet Rosetta

	(c)Copyright 2017-2020  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

******************************************************************************************

The RosettaFioUtils.c file contains the basic OS and app independent support functions.

The RosettaFioUtils.h file contains the external declarations and definitions.

Content of RosettaFioUtils.c
1.) Date and time functions
2.) String to and to String functions 
3.) String UTF8 cmp functions
4.) String functions
5.) File functions
6.) Xml Syntax functions
7.) Link LIst functions

**********************************************************************************/
#ifdef  _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
//#define   malloc(s)				_malloc_dbg(s, _NORMAL_BLOCK, __FILE__, __LINE__)
#define   calloc(_Count, _Size)		_calloc_dbg(_Count, _Size, _NORMAL_BLOCK, __FILE__, __LINE__)
// etc...
#endif
#include <stdlib.h>
#include <memory.h>

#ifdef WIN32
#include <BaseTsd.h>
#ifndef BOOL
#define BOOL		int
#endif
#elif  IOS
#define BOOL		int
#elif  OSX
#define BOOL		int
#elif  ANDROID
typedef int			BOOL;
#else
#ifndef BOOL
#define BOOL		int
#endif
#endif

#define RET_ERR		0	// Error
#define RET_OK		1	// OK

#define WRITE_BUFFER_LENGTH			10240
#define TEMPSTRING_LENGTH			1024

extern char NulString[1];

#define ERR_OK						0
#define ERR_FILE_NAME				(-1)
#define ERR_FILE_CONTENT			(-2)
#define ERR_EXPIRED					(-3)
#define ERR_OS_VER					(-4)
#define ERR_APP_REJECT				(-80)
#define ERR_CLOSE_FILE				(-97)
#define ERR_OPEN_FILE				(-98)
#define ERR_TECHNICAL				(-99)

// RosettaFioTypes.h
#define FILE_DEFAULT				((BITMASK)0x0000)
#define FILE_ACT_MASK				((BITMASK)0x1000)
#define FILE_TYPE_MASK				((BITMASK)0x0f00)
#define CONTENT_MASK				((BITMASK)0x00ff)
#define FILE_ACT_OPEN				((BITMASK)0x0000)
#define FILE_ACT_SAVE				((BITMASK)0x1000)
#define FILE_ACT_IMPORT				((BITMASK)0x2000)
#define FILE_TYPE_STD				((BITMASK)0x0000)
#define FILE_TYPE_TXT				((BITMASK)0x0100)
#define FILE_TYPE_CSV				((BITMASK)0x0200)
#define FILE_TYPE_PDF				((BITMASK)0x0300)
#define FILE_TYPE_TITLE				((BITMASK)0x0400)
#define FILE_TYPE_PRINT				((BITMASK)0x0500)

#define DOUBLE_PRECITION			15 // MSC double specification

#define NORMAL_MONTHS_PER_YEAR		12
#define MAX_MONTH_PER_YEAR			18
#define MIN_MONTH_PER_YEAR			6
#define CALENDAR_YEAR_START_MONTH	1

#define LENGTH_NUMERIC_STR_MAX		22 //(Maxlenth for a signed int64, turns over the edge else)
#define LENGTH_AMOUNT_STR_MAX		19 //(Maxlenth for a signed long long, turns over the edge else)
#define LENGTH_AMOUNT_STRING		(LENGTH_AMOUNT_STR_MAX+2) //(LENGTH_AMOUNT_STR_MAX+one comma and one sign-sign)
									// LENGTH_AMOUNT_SPACED_STR is similar to long long with speces and signed, limits by the atof function, in Win32 about 14 digits
#define LENGTH_AMOUNT_SPACED_STR	(LENGTH_AMOUNT_STR_MAX+10) //(LENGTH_AMOUNT_STR_MAX+one comma and one sign-sign + spaces) 
#define LENGTH_QUANT_STRING			29 
#define LENGTH_LONG_DATE			13

#ifndef UCHAR
typedef unsigned char UCHAR;
#endif

#ifdef WIN32
#include <BaseTsd.h>
#ifndef BOOL
#define BOOL  int
#endif
#elif  ANDROID
typedef int 						BOOL;
typedef long long int				INT64;
typedef unsigned long long int		UINT64;
#else
#define INT64 int64_t
#endif

#ifndef FALSE
#define FALSE						0
#endif
#ifndef TRUE
#define TRUE						1
#endif

#define MONTH_LENGTH				100
#define QUARTER_LENGTH				300
#define TERTIAL_LENGTH				400
#define SEMI_LENGTH					600
#define ANNUAL_LENGTH				10000
#define MONTH						'M'
#define QUARTER						'Q'		
#define TERTIAL						'T'
#define SEMI						'S'
#define ANNUAL						'A'
#define MONTHPERIOD					"M1"
#define QUARTERPERIOD				"Q1"		
#define TERTIALPERIOD				"T1"
#define SEMIPERIOD					"S1"
#define ANNUALPERIOD				"A1"


typedef struct
{
	char *pLocaleLanguage;
	char *pLocaleCountry;
	char *pLocaleCurrency;
	char *pLocaleShortDateTimePattern;
	char *pLocaleDecimalSeparator;
	char *pLocaleThousandSeparator;
	char *pLocaleThousandGrouping;
	char *pLocaleNegativePrefix;
	char *pLocaleNegativeSuffix;
	char *pMonth[NORMAL_MONTHS_PER_YEAR];
	char *pOperator;
} LocaleDataType;

typedef struct {
	char *pTag;
	char *pData;
	char *Attr;
	char *pBuffer;
} XmlTagDataType;

typedef struct {
	void *vpData;
	void *vpNextRecord;
	void *vpPrevRecord;
} LLISTITEMTYPE;

typedef struct {
	LLISTITEMTYPE *FirstRecord;
	LLISTITEMTYPE *CursorRecord;
	LLISTITEMTYPE *LastRecord;
} LLISTTYPE;

#define LLIST	LLISTTYPE *

// Date and time functions
char *GetSystemYear(void);
DATEINT GetSystemDate(void);
DATETIMEINT GetSystemDateTime(void);
BOOL IsLeapYr(int iYear);
DATEINT GetFirstDateInYear(DATEINT Date);
DATEINT GetNextMonthLastDate(DATEINT Date);
DATEINT GetPrevMonthLastDate(DATEINT Date);
DATEINT GetNextDate(DATEINT Date);
DATEINT GetPrevDate(DATEINT Date);
DATEINT GetPeriodDate(DATEINT BalancePeriodDate, char *pPeriodText);
DATEINT GetLastDateInPeriod(DATEINT Date, char *pPeriodLabel);
DATEINT GetLastDateInMonth(DATEINT Date);
BOOL CheckDateTime(DATEINT Date, TIMEINT Time);
DATEINT AddDays(DATEINT BaseDate, int iDays);
DATEINT AddMonths(DATEINT BaseDate, int iMonths);
DATEINT AddYears(DATEINT BaseDate, int iYears);

// String to and to String functions
BOOL DecimalCmp(QUANT Decimal1, QUANT Decimal2);
int Strtoi(char const* String);
long Strtol(char const* String);
long long Strtoll(char const* String);
unsigned int Strtoui(char const* String);
unsigned long Strtoul(char const* String);
unsigned long long Strtoull(char const* String);
char *itoStr(int Value, char *pString);
char *ltoStr(long Value, char *pString);
char *lltoStr(long long Value, char *pString);
char *utoStr(unsigned int Value, char *pString, int Radix);
char *ultoStr(unsigned long Value, char *pString, int Radix);
char *ulltoStr(unsigned long long Value, char *pString, int Radix);
UCHAR *StrUpper(UCHAR *string);
UCHAR *StrLwr(UCHAR *string);
BOOL StringToBool(const char *pString);
char *BoolToStr(BOOL Bool, char *pString);
NUMERIC StringToNumeric(const char *pString);
char *NumericToStr(NUMERIC Numeric, char *String);
COUNTER StringToCounter(const char *pString);
char *CounterToStr(COUNTER Numeric, char *String);
void *StringToPtr(const char *pString);
char *PtrtoStr(void *Value, char *pString);
TIMEINT StringToTimeLocale(const char *TimeString, LocaleDataType *pLocaleData);
TIMEINT StringToTime(const char *TimeString);
DATEINT StringToDateLocale(const char *pDateString, LocaleDataType *pLocaleData);
DATEINT StringToDate(const char *pDateString);
DATETIMEINT StringToDateTimeLocale(const char *DateTimeString, LocaleDataType *pLocaleData);
DATETIMEINT StringToDateTime(const char *DateTimeString);
//static char *ConvertDecimalExponentials(char *pDecimalString);
AMOUNT StringToAmountLocale(const char *pAmountString, LocaleDataType *pLocaleData);
AMOUNT StringToAmount(const char *pAmountString);
QUANT StringToDecimalLocale(const char *pDecimalString, LocaleDataType *pLocaleData);
QUANT StringToDecimal(const char *pDecimalString);
char *TimeToStrLocale(TIMEINT Time, char *TimeString, LocaleDataType *pLocaleData);
char *TimeToStr(TIMEINT Time, char *TimeString);
char *DateToStrLocale(DATEINT Date, char *pDateString, LocaleDataType *pLocaleData);
char *DateToStr(DATEINT Date, char *pDateString);
char *DateTimeToStrLocale(DATETIMEINT DateTime, char *DateTimeString, LocaleDataType *pLocaleData);
char *DateTimeToStr(DATETIMEINT DateTime, char *DateTimeString);
char *AmountToStrLocale(AMOUNT Amount, char *pAmountString, LocaleDataType *pLocaleData);
char *AmountToStr(AMOUNT Amount, char *pAmountString);
char *DecimalToStrLocale(QUANT Decimal, char *pDecimalString, LocaleDataType *pLocaleData);
char *DecimalToStr(QUANT Decimal, char *pDecimalString);
char *MakeLocalNumericString(char *pAmountString, LocaleDataType *pLocaleData);

// String UTF8 cmp functions
unsigned char *StrLwrExt(unsigned char *pString);
int StrnCiCmp(const char *s1, const char *s2, size_t ztCount);
int StrCiCmp(const char *s1, const char *s2);
char *StrCiStr(const char *s1, const char *s2);

// String functions
char *StringUpdate(char **ppOriginalData, char *pNewData);
char *CopyString(char **ppTarget, const char *pData);
char *FreeString(char **ppData);
void ReadString(char **ppData, char *pRead);

// File functions
FILE *FileOpen(char *pFileName, char *pMode);
char *FileRead(char *szFileName);
size_t FileWrite(char *szFileName, char *pBuffer);
int ufWrite(FILE *pStream, char *pBuffer);

// Xml Syntax functions
size_t xmlWrite(FILE *pStream, char *pBuffer);
int xmlWriteLine(FILE *pStream, int iTab, char *attr, char *pPrefix, char *pData, char *pSuffix);
XmlTagDataType *GetNextXmlTag(XmlTagDataType *XmlTagData);
char *CloseXmlAttrString(char *pText, char *pTag);
char *ClearXmlEsc(char *pBuffer);
char *ReadXmlDataString(char *Indata);

// System time
char *GetSystemUser(char **ppTarget);
DATEINT GetSystemDate(void);


/*********************************************************************

Linked list functions

The linked list is based on the struct LLISTITEMTYPE that contains a void pointer to the Data
and reference to the Next Record and the PreviousRecord.

Three of these forms the LLISTTYPE set of the First Record, the Cursor Record and the Last Record.
This way there are entrances and ability to wind the linked list back and forward to retrieve data.

The functions in LList.c are to handle the linked list database system.

************************************************************************/
LLIST LListInit(void);
BOOL LListDestroy(LLIST pLList);
BOOL LListClear(LLIST pLList);
BOOL LListDeleteData(LLIST pLList);
BOOL LListErase(LLIST pLList);
BOOL LListAddRecord(LLIST pLList, void *vpData);
BOOL LListUpdateDataCursorRecord(LLIST pLList, void *vpNewData);
void *LListSearchDeleteRecord(LLIST pLList, void *vpSearchData);
void **LListSearchData(LLIST pLList, void *vpSearchData);
void **LListSearchText(LLIST pLList, char *szSearchText);
void LListNulDataCursorRecord(LLIST pLList);
void *LListDeleteCursorRecord(LLIST pLList);
void *LListEraseCursorRecord(LLIST pLList);
void *LListGetCursorRecordData(LLIST pLList);
void *LListGetNextRecordData(LLIST pLList);
void *LListGetPreviousRecordData(LLIST pLList);
void *LListGetFirstRecordData(LLIST pLList);
void *LListGetLastRecordData(LLIST pLList);
BOOL LListMoveCursorForward(LLIST pLList);
BOOL LListMoveCursorBackward(LLIST pLList);
BOOL LListMoveCursorToFirst(LLIST pLList);
BOOL LListMoveCursorToLast(LLIST pLList);
LLIST LListCopyLList(LLIST *ppTargetLList, LLIST pOriginalLList);
LLIST LListCopyLListWithContent(LLIST pTargetLList, LLIST pOriginalLList, int DataSize);
unsigned long long LListGetListSize(LLIST pLList);
void *LListGetCursorRecordPtr(LLIST pLList);
