/*********************************************************************************

	RosettaFioXbrlGl.h

	Alphabet Rosetta

	(c)Copyright 2017-2020  Alphabet AB, Sweden (Jan Bergstrom)

	http://www.alphabet.se

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.

	We ask for attribution to us, when used.

******************************************************************************************

The RosettaFioXbrlGl.c file contains functions to interpret and write XBRL/GL-files 
(www.xbrl.org/the-standard/what/global-ledger/) into the Alphabet Rosetta Accounting 
basic data logics and database system.

The RosettaFioXbrlGl.h file contains the external declarations and definitions.

Content of RosettaFioXbrlGl.c
1.) Interpret XBRL/GL file data
2.) Create XBRL/GL file data

**********************************************************************************/

// Interpret XBRL/GL file data
void *OpenXbrlGlFile(char*pFilePathName, LLIST plErrorList);
void *InterpretXbrlGl(char *pBuffer, LLIST plErrorList);

// Create XBRL/GL file data
int FileWriteXbrlGl(
	char *pFilePathName,
	void *vpAccounting,
	char *pContent,
	char *pGenUserName,
	char *pStylesheetFilePathName,
	BOOL ChartsOption,
	BOOL ObjectOption,
	BOOL BalanceOption,
	BOOL PeriodOption,
	BOOL VouchersOption,
	BOOL QuantityOption,
	BOOL LeanChartsOption,
	BOOL CompactFileOption,
	BOOL ExtendedOption,
	char *pInitialDate,
	char *pFinalDate,
	char *FilterEnteredBy,
	char *pProgramName,
	char *pProgramVersion,
	LocaleDataType *pLocaleData);

int StreamWriteXbrlGl(
	FILE *pStream,
	void *pAccounting,
	char *pContent,
	char *pGenUserName,
	char *pStylesheetFilePathName,
	BOOL ChartsOption,
	BOOL ObjectOption,
	BOOL BalanceOption,
	BOOL PeriodOption,
	BOOL VouchersOption,
	BOOL QuantityOption,
	BOOL LeanChartsOption,
	BOOL CompactFileOption,
	BOOL ExtendedOption,
	char *pInitialDate,
	char *pFinalDate,
	char *FilterEnteredBy,
	char *pProgramName,
	char *pProgramVersion,
	LocaleDataType *pLocaleData);

#define ERR_MEMORY					-1
#define ERR_FATAL					-2
#define ERR_SIE_TYPE				-11
#define ERR_SIE_PERIOD				-12
#define ERR_SIE_COMPANY_NAME		-13

#define ACCOUNT_TYPE_TOTAL			"Total"
#define ACCOUNT_TYPE_BALANCE		"Balance"
#define ACCOUNT_TYPE_RESULT			"Result"
#define ACCOUNT_TYPE_ACTIVE			"Active"
#define ACCOUNT_TYPE_PASSIVE		"Passive"
#define ACCOUNT_TYPE_PROFIT			"Profit"
#define ACCOUNT_TYPE_EARNING		"Earning"
#define ACCOUNT_TYPE_OUTCOME		"Outcome"
#define ACCOUNT_TYPE_TANGIBLE		"Tangible"
#define ACCOUNT_TYPE_STOCK			"Stock"
#define ACCOUNT_TYPE_RECEIVABLES	"Receivables"
#define ACCOUNT_TYPE_ASSETS			"Assets"
#define ACCOUNT_TYPE_EQUITIES		"Equities"
#define ACCOUNT_TYPE_PROVISIONS		"Provisions"
#define ACCOUNT_TYPE_LIABILITIES	"Liabilities"
#define ACCOUNT_TYPE_REVENUES		"Revenues"
#define ACCOUNT_TYPE_SUPPLIES		"Supplies"
#define ACCOUNT_TYPE_EXPENSES		"Expenses"
#define ACCOUNT_TYPE_STAFF			"Staff"
#define ACCOUNT_TYPE_LOSS_GAINS		"LossGains"
#define ACCOUNT_TYPE_FINANCIAL		"Financial"
#define ACCOUNT_TYPE_AUX			"Auxiliary"

